from __future__ import annotations

from pathlib import Path

import pytest

from kanbus import wiki
from kanbus import config_loader, project
from kanbus.console_snapshot import ConsoleSnapshotError
from kanbus.config_loader import ConfigurationError
from kanbus.project import ProjectMarkerError

from test_helpers import build_issue, build_project_configuration


def test_get_string_and_serialize_issue() -> None:
    assert wiki._get_string(None) is None
    assert wiki._get_string("x") == "x"
    with pytest.raises(wiki.WikiError, match="invalid query parameter"):
        wiki._get_string(123)

    issue = build_issue("kanbus-1")
    payload = wiki._serialize_issue(issue)
    assert payload["id"] == "kanbus-1"
    assert payload["type"] == issue.issue_type


def test_wiki_context_query_count_issue_and_invalid_sort() -> None:
    a = build_issue("kanbus-a", title="B title", priority=3, status="open")
    b = build_issue("kanbus-b", title="A title", priority=1, status="open")
    c = build_issue("kanbus-c", title="C title", priority=2, status="closed")
    context = wiki.WikiContext([a, b, c])

    base = context.query(status="open")
    assert {row["id"] for row in base} == {"kanbus-a", "kanbus-b"}

    by_title = context.query(status="open", sort="title")
    assert [row["id"] for row in by_title] == ["kanbus-b", "kanbus-a"]

    by_priority = context.query(status="open", sort="priority")
    assert [row["id"] for row in by_priority] == ["kanbus-b", "kanbus-a"]

    assert context.count(status="closed") == 1
    assert context.issue("kanbus-a")["id"] == "kanbus-a"  # type: ignore[index]
    assert context.issue("missing") is None

    with pytest.raises(wiki.WikiError, match="invalid sort key"):
        context.query(sort="bad")


def test_wiki_render_cache_helpers(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    page = tmp_path / "page.md"
    page.write_text("hello", encoding="utf-8")
    issues = [build_issue("kanbus-1")]

    key = wiki._wiki_render_cache_key(page, issues)
    assert len(key) == 64

    cache_dir = tmp_path / "cache"
    assert wiki._wiki_render_read_cache(cache_dir, "missing") is None

    wiki._wiki_render_write_cache(cache_dir, "k1", "content")
    assert wiki._wiki_render_read_cache(cache_dir, "k1") == "content"

    # Read errors should return None.
    monkeypatch.setattr(
        Path,
        "read_text",
        lambda *_args, **_kwargs: (_ for _ in ()).throw(OSError("boom")),
    )
    assert wiki._wiki_render_read_cache(cache_dir, "k1") is None


def test_wiki_render_log_cache_hit(tmp_path: Path) -> None:
    cache_dir = tmp_path / ".cache" / "wiki_render"
    wiki._wiki_render_log_cache_hit(cache_dir)
    log = cache_dir.parent / "wiki_cache_hits.log"
    assert log.read_text(encoding="utf-8") == "1\n"


def test_load_ai_config_and_project_dir_success_and_failures(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    cfg = build_project_configuration().model_copy(
        update={"project_directory": "project"}
    )
    cfg_path = tmp_path / ".kanbus.yml"

    monkeypatch.setattr(project, "get_configuration_path", lambda _root: cfg_path)
    monkeypatch.setattr(config_loader, "load_project_configuration", lambda _path: cfg)

    ai_config, project_dir = wiki._load_ai_config_and_project_dir(tmp_path)
    assert ai_config is None
    assert project_dir == "project"

    monkeypatch.setattr(
        project,
        "get_configuration_path",
        lambda _root: (_ for _ in ()).throw(ProjectMarkerError("missing")),
    )
    assert wiki._load_ai_config_and_project_dir(tmp_path) == (None, None)

    monkeypatch.setattr(project, "get_configuration_path", lambda _root: cfg_path)
    monkeypatch.setattr(
        config_loader,
        "load_project_configuration",
        lambda _path: (_ for _ in ()).throw(ConfigurationError("bad")),
    )
    assert wiki._load_ai_config_and_project_dir(tmp_path) == (None, None)


def test_render_template_string_success_and_errors() -> None:
    issues = [build_issue("kanbus-1", title="Hello")]

    rendered = wiki.render_template_string("{{ issue('kanbus-1').title }}", issues)
    assert rendered == "Hello"

    with pytest.raises(wiki.WikiError, match="invalid query parameter"):
        wiki.render_template_string("{{ count(status=1) }}", issues)

    with pytest.raises(wiki.WikiError):
        wiki.render_template_string("{% for x in %}", issues)


def test_render_wiki_page_raises_for_missing_page(tmp_path: Path) -> None:
    request = wiki.WikiRenderRequest(
        root=tmp_path, page_path=Path("project/wiki/missing.md")
    )
    with pytest.raises(wiki.WikiError, match="wiki page not found"):
        wiki.render_wiki_page(request)


def test_render_wiki_page_wraps_console_snapshot_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    page = tmp_path / "project" / "wiki" / "p.md"
    page.parent.mkdir(parents=True)
    page.write_text("x", encoding="utf-8")

    monkeypatch.setattr(
        wiki,
        "get_issues_for_root",
        lambda _root: (_ for _ in ()).throw(ConsoleSnapshotError("snapshot failed")),
    )

    with pytest.raises(wiki.WikiError, match="snapshot failed"):
        wiki.render_wiki_page(
            wiki.WikiRenderRequest(root=tmp_path, page_path=Path("project/wiki/p.md"))
        )


def test_render_wiki_page_cache_hit_returns_cached_content(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    page = tmp_path / "project" / "wiki" / "p.md"
    page.parent.mkdir(parents=True)
    page.write_text("{{ 1 }}", encoding="utf-8")

    monkeypatch.setattr(
        wiki, "get_issues_for_root", lambda _root: [build_issue("kanbus-1")]
    )
    monkeypatch.setattr(
        wiki, "_load_ai_config_and_project_dir", lambda _root: (None, "project")
    )
    monkeypatch.setattr(wiki, "_wiki_render_cache_key", lambda _p, _issues: "k")
    monkeypatch.setattr(wiki, "_wiki_render_read_cache", lambda _dir, _key: "cached")

    logged: list[str] = []
    monkeypatch.setattr(
        wiki, "_wiki_render_log_cache_hit", lambda _dir: logged.append("hit")
    )

    rendered = wiki.render_wiki_page(
        wiki.WikiRenderRequest(root=tmp_path, page_path=Path("project/wiki/p.md"))
    )
    assert rendered == "cached"
    assert logged == ["hit"]


def test_render_wiki_page_renders_and_writes_cache(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    page = tmp_path / "project" / "wiki" / "p.md"
    page.parent.mkdir(parents=True)
    page.write_text("Count={{ count(status='open') }}", encoding="utf-8")

    issues = [
        build_issue("kanbus-1", status="open"),
        build_issue("kanbus-2", status="closed"),
    ]
    monkeypatch.setattr(wiki, "get_issues_for_root", lambda _root: issues)
    monkeypatch.setattr(
        wiki, "_load_ai_config_and_project_dir", lambda _root: (None, "project")
    )
    monkeypatch.setattr(wiki, "_wiki_render_cache_key", lambda _p, _issues: "k")
    monkeypatch.setattr(wiki, "_wiki_render_read_cache", lambda _dir, _key: None)
    monkeypatch.setattr(
        wiki,
        "make_ai_summarize",
        lambda *_args, **_kwargs: (lambda *_a, **_k: "summary"),
    )
    monkeypatch.chdir(tmp_path)

    writes: list[str] = []

    def _write(_cache_dir: Path, _key: str, content: str) -> None:
        writes.append(content)

    monkeypatch.setattr(wiki, "_wiki_render_write_cache", _write)

    rendered = wiki.render_wiki_page(
        wiki.WikiRenderRequest(root=tmp_path, page_path=Path("project/wiki/p.md"))
    )
    assert rendered == "Count=1"
    assert writes == ["Count=1"]


def test_render_wiki_page_wraps_template_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    page = tmp_path / "project" / "wiki" / "p.md"
    page.parent.mkdir(parents=True)
    page.write_text("{% for x in %}", encoding="utf-8")

    monkeypatch.setattr(
        wiki, "get_issues_for_root", lambda _root: [build_issue("kanbus-1")]
    )
    monkeypatch.setattr(
        wiki, "_load_ai_config_and_project_dir", lambda _root: (None, "project")
    )
    monkeypatch.setattr(wiki, "_wiki_render_cache_key", lambda _p, _issues: "k")
    monkeypatch.setattr(wiki, "_wiki_render_read_cache", lambda _dir, _key: None)
    monkeypatch.setattr(
        wiki,
        "make_ai_summarize",
        lambda *_args, **_kwargs: (lambda *_a, **_k: "summary"),
    )
    monkeypatch.chdir(tmp_path)

    with pytest.raises(wiki.WikiError):
        wiki.render_wiki_page(
            wiki.WikiRenderRequest(root=tmp_path, page_path=Path("project/wiki/p.md"))
        )


def test_render_wiki_page_re_raises_wiki_error_from_template_context(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    page = tmp_path / "project" / "wiki" / "p.md"
    page.parent.mkdir(parents=True)
    page.write_text("{{ count(status=1) }}", encoding="utf-8")

    monkeypatch.setattr(
        wiki, "get_issues_for_root", lambda _root: [build_issue("kanbus-1")]
    )
    monkeypatch.setattr(
        wiki, "_load_ai_config_and_project_dir", lambda _root: (None, "project")
    )
    monkeypatch.setattr(wiki, "_wiki_render_cache_key", lambda _p, _issues: "k")
    monkeypatch.setattr(wiki, "_wiki_render_read_cache", lambda _dir, _key: None)
    monkeypatch.setattr(
        wiki,
        "make_ai_summarize",
        lambda *_args, **_kwargs: (lambda *_a, **_k: "summary"),
    )
    monkeypatch.chdir(tmp_path)

    with pytest.raises(wiki.WikiError, match="invalid query parameter"):
        wiki.render_wiki_page(
            wiki.WikiRenderRequest(root=tmp_path, page_path=Path("project/wiki/p.md"))
        )


def test_list_wiki_pages_success_absolute_relative_and_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    cfg_path = tmp_path / ".kanbus.yml"

    monkeypatch.setattr(project, "get_configuration_path", lambda _root: cfg_path)

    cfg = build_project_configuration().model_copy(
        update={"project_directory": "project", "wiki_directory": "wiki"}
    )
    monkeypatch.setattr(config_loader, "load_project_configuration", lambda _path: cfg)

    root_wiki = tmp_path / "project" / "wiki"
    (root_wiki / "sub").mkdir(parents=True)
    (root_wiki / "a.md").write_text("a", encoding="utf-8")
    (root_wiki / "sub" / "b.md").write_text("b", encoding="utf-8")
    (root_wiki / "skip.txt").write_text("x", encoding="utf-8")

    paths = wiki.list_wiki_pages(tmp_path)
    assert paths == ["project/wiki/a.md", "project/wiki/sub/b.md"]

    cfg_outside = build_project_configuration().model_copy(
        update={"project_directory": "project", "wiki_directory": "../docs/wiki"}
    )
    monkeypatch.setattr(
        config_loader, "load_project_configuration", lambda _path: cfg_outside
    )

    outside_wiki = tmp_path / "docs" / "wiki"
    outside_wiki.mkdir(parents=True)
    (outside_wiki / "c.md").write_text("c", encoding="utf-8")

    paths2 = wiki.list_wiki_pages(tmp_path)
    assert paths2 == ["docs/wiki/c.md"]

    # Missing directory returns empty list.
    cfg_missing = build_project_configuration().model_copy(
        update={"project_directory": "project", "wiki_directory": "wiki-missing"}
    )
    monkeypatch.setattr(
        config_loader, "load_project_configuration", lambda _path: cfg_missing
    )
    assert wiki.list_wiki_pages(tmp_path) == []

    monkeypatch.setattr(
        project,
        "get_configuration_path",
        lambda _root: (_ for _ in ()).throw(ProjectMarkerError("missing")),
    )
    with pytest.raises(wiki.WikiError, match="missing"):
        wiki.list_wiki_pages(tmp_path)

    monkeypatch.setattr(project, "get_configuration_path", lambda _root: cfg_path)
    monkeypatch.setattr(
        config_loader,
        "load_project_configuration",
        lambda _path: (_ for _ in ()).throw(ConfigurationError("bad config")),
    )
    with pytest.raises(wiki.WikiError, match="bad config"):
        wiki.list_wiki_pages(tmp_path)



from __future__ import annotations

from datetime import datetime, timezone
from types import SimpleNamespace

import pytest

from kanbus import issue_display, issue_line, wiki
from kanbus.models import (
    CategoryDefinition,
    DependencyLink,
    IssueComment,
    StatusDefinition,
)

from test_helpers import build_issue, build_project_configuration


def _comment(
    text: str, author: str = "dev", comment_id: str | None = None
) -> IssueComment:
    return IssueComment.model_validate(
        {
            "id": comment_id,
            "author": author,
            "text": text,
            "created_at": datetime(2026, 3, 9, tzinfo=timezone.utc).isoformat(),
        }
    )


def test_issue_display_color_helpers(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(
        issue_display.click,
        "get_current_context",
        lambda silent=True: SimpleNamespace(color=True),
    )
    assert issue_display._should_use_color() is True

    monkeypatch.setattr(
        issue_display.click, "get_current_context", lambda silent=True: None
    )
    monkeypatch.setattr(issue_display.sys.stdout, "isatty", lambda: True)
    monkeypatch.delenv("NO_COLOR", raising=False)
    assert issue_display._should_use_color() is True

    monkeypatch.setenv("NO_COLOR", "1")
    assert issue_display._should_use_color() is False

    monkeypatch.setattr(
        issue_display.click, "style", lambda text, fg=None: f"[{fg}]{text}"
    )
    assert issue_display._dim("ID:", use_color=False) == "ID:"
    assert issue_display._dim("ID:", use_color=True) == "[bright_black]ID:"
    assert issue_display._normalize_color("red") == "red"
    assert issue_display._normalize_color("not-a-color") is None
    assert issue_display._paint("x", "red", use_color=False) == "x"
    assert issue_display._paint("x", "bad", use_color=True) == "x"
    assert issue_display._paint("x", "red", use_color=True) == "[red]x"


def test_render_description_and_comments_success(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    issue = build_issue("kanbus-1", title="T", status="open")
    issue.description = "{{ desc }}"
    issue.comments = [_comment("{{ c1 }}"), _comment("{{ c2 }}")]
    monkeypatch.setattr(
        wiki, "render_template_string", lambda value, _issues: f"R:{value}"
    )

    description, comments = issue_display._render_description_and_comments(
        issue, [issue]
    )

    assert description == "R:{{ desc }}"
    assert comments == ["R:{{ c1 }}", "R:{{ c2 }}"]


def test_render_description_and_comments_wiki_error_fallback(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    issue = build_issue("kanbus-1", title="T", status="open")
    issue.description = "raw description"
    issue.comments = [_comment("raw comment")]

    def fail(_value: str, _issues: list[object]) -> str:
        raise wiki.WikiError("boom")

    monkeypatch.setattr(wiki, "render_template_string", fail)

    description, comments = issue_display._render_description_and_comments(
        issue, [issue]
    )
    assert description == "raw description"
    assert comments == ["raw comment"]


def test_render_description_and_comments_empty_description(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    issue = build_issue("kanbus-1")
    issue.description = ""
    issue.comments = [_comment("comment text")]
    monkeypatch.setattr(
        wiki, "render_template_string", lambda value, _issues: f"R:{value}"
    )
    description, comments = issue_display._render_description_and_comments(
        issue, [issue]
    )
    assert description == ""
    assert comments == ["R:comment text"]


def test_format_issue_for_display_sections_and_context(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    configuration = build_project_configuration()
    configuration.type_colors["task"] = "green"
    configuration.categories = [
        CategoryDefinition.model_validate({"name": "To do", "color": "magenta"})
    ]
    configuration.statuses = [
        StatusDefinition.model_validate(
            {"key": "open", "name": "Open", "category": "To do"}
        )
    ]
    configuration.priorities[2].color = "bright_blue"

    issue = build_issue("kanbus-1", title="My Title", status="open", issue_type="task")
    issue.description = "desc"
    issue.dependencies = [
        DependencyLink.model_validate({"target": "kanbus-2", "type": "blocked-by"})
    ]
    issue.comments = [
        _comment("first", comment_id="abcdef123456"),
        SimpleNamespace(id=None, author="", text="second"),
    ]

    monkeypatch.setattr(
        issue_display, "format_issue_key", lambda identifier, _ctx: f"FMT:{identifier}"
    )
    monkeypatch.setattr(
        issue_display,
        "_render_description_and_comments",
        lambda _issue, _all: ("rendered desc", ["rc1", "rc2"]),
    )

    rendered = issue_display.format_issue_for_display(
        issue,
        configuration=configuration,
        use_color=False,
        project_context=True,
        all_issues=[issue],
    )

    assert "ID: FMT:kanbus-1" in rendered
    assert "Title: My Title" in rendered
    assert "Description:" in rendered
    assert "rendered desc" in rendered
    assert "Dependencies:" in rendered
    assert "blocked-by: kanbus-2" in rendered
    assert "Comments:" in rendered
    assert "[abcdef] dev: rc1" in rendered
    assert "unknown: rc2" in rendered


def test_format_issue_for_display_with_colorized_muted_fields(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    issue = build_issue("kanbus-9", parent=None, labels=[])
    issue.assignee = None
    monkeypatch.setattr(
        issue_display, "format_issue_key", lambda identifier, _ctx: identifier
    )
    monkeypatch.setattr(
        issue_display.click, "style", lambda text, fg=None: f"[{fg}]{text}"
    )

    rendered = issue_display.format_issue_for_display(issue, use_color=True)
    assert "[bright_black]-" in rendered


def test_format_issue_for_display_uses_explicit_status_color() -> None:
    configuration = build_project_configuration()
    configuration.categories = [
        CategoryDefinition.model_validate({"name": "To do", "color": "blue"})
    ]
    configuration.statuses = [
        StatusDefinition.model_validate(
            {"key": "open", "name": "Open", "category": "To do", "color": "red"}
        )
    ]
    issue = build_issue("kanbus-1", status="open")
    rendered = issue_display.format_issue_for_display(
        issue, configuration=configuration, use_color=False
    )
    assert "Status: open" in rendered


def test_issue_line_color_resolution_helpers() -> None:
    configuration = build_project_configuration()
    configuration.categories = [
        CategoryDefinition.model_validate({"name": "To do", "color": "red"})
    ]
    configuration.statuses = [
        StatusDefinition.model_validate(
            {"key": "open", "name": "Open", "category": "To do"}
        )
    ]
    configuration.statuses[0].color = "red"

    assert issue_line._normalize_cli_color("gray") == "bright_black"
    assert issue_line._normalize_cli_color("grey") == "bright_black"
    assert issue_line._normalize_cli_color("not-a-color") is None

    assert issue_line._resolve_status_color("open", configuration) == "red"
    assert issue_line._resolve_status_color("missing", configuration) == "white"
    configuration.statuses[0].color = None
    configuration.categories[0].color = "cyan"
    assert issue_line._resolve_status_color("open", configuration) == "cyan"

    configuration.priorities[2].color = "magenta"
    assert issue_line._resolve_priority_color(2, configuration) == "magenta"
    assert issue_line._resolve_priority_color(4, None) == "white"

    configuration.type_colors["task"] = "green"
    assert issue_line._resolve_type_color("task", configuration) == "green"
    assert issue_line._resolve_type_color("unknown", None) == "white"


def test_issue_line_safe_color_and_widths(monkeypatch: pytest.MonkeyPatch) -> None:
    assert issue_line._safe_color(lambda t, fg=None: f"[{fg}]{t}", "x", None) == "x"
    assert (
        issue_line._safe_color(lambda t, fg=None: f"[{fg}]{t}", "x", "red") == "[red]x"
    )

    one = build_issue("kanbus-1", issue_type="task", status="open")
    two = build_issue(
        "kanbus-22", issue_type="bug", status="in_progress", parent="kanbus-1"
    )
    widths = issue_line.compute_widths([one, two])
    assert widths["identifier"] >= len("kanbus-22")
    assert widths["status"] >= len("in_progress")

    monkeypatch.setattr(
        issue_line,
        "format_issue_key",
        lambda identifier, project_context=False: (
            identifier.replace("kanbus-", "k-") if project_context else identifier
        ),
    )
    formatted = issue_line.format_issue_line(one, porcelain=True, project_context=True)
    assert " | " in formatted
    assert "k-1" in formatted


def test_issue_line_format_non_porcelain_with_color_paths(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    issue = build_issue("kanbus-1", issue_type="task", status="open", parent=None)
    issue.custom["project_path"] = "alpha/project"
    configuration = build_project_configuration()
    configuration.type_colors["task"] = "green"
    configuration.priorities[2].color = "blue"
    configuration.categories = [
        CategoryDefinition.model_validate({"name": "To do", "color": "cyan"})
    ]
    configuration.statuses = [
        StatusDefinition.model_validate(
            {"key": "open", "name": "Open", "category": "To do"}
        )
    ]
    configuration.statuses[0].color = "cyan"

    monkeypatch.setattr(issue_line.sys.stdout, "isatty", lambda: True)
    monkeypatch.delenv("NO_COLOR", raising=False)
    monkeypatch.setattr(
        issue_line,
        "format_issue_key",
        lambda identifier, project_context=False: identifier,
    )

    line = issue_line.format_issue_line(
        issue,
        porcelain=False,
        colorizer=lambda text, fg=None: f"[{fg}]{text}",
        widths={"type": 1, "identifier": 8, "parent": 1, "status": 4, "priority": 2},
        configuration=configuration,
        use_color=None,
    )

    assert line.startswith("alpha/project ")
    assert "[green]T" in line
    assert "[cyan]open" in line
    assert "[blue]P2" in line


def test_issue_line_no_color_branch(monkeypatch: pytest.MonkeyPatch) -> None:
    issue = build_issue("kanbus-1", issue_type="task", status="open")
    monkeypatch.setenv("NO_COLOR", "1")
    monkeypatch.setattr(issue_line.sys.stdout, "isatty", lambda: True)

    line = issue_line.format_issue_line(issue, porcelain=False, use_color=None)
    assert "[" not in line



from __future__ import annotations

from datetime import datetime, timedelta, timezone
from pathlib import Path
import json
import subprocess
import tempfile
from types import SimpleNamespace

import kanbus.overlay as overlay_module
from kanbus.models import IssueData, OverlayConfig
from kanbus.overlay import (
    OverlayIssueRecord,
    gc_overlay,
    install_overlay_hooks,
    reconcile_overlay,
    overlay_issue_path,
    resolve_issue_with_overlay,
    write_overlay_issue,
)


def _issue(identifier: str, updated_at: datetime) -> IssueData:
    return IssueData(
        id=identifier,
        title="Overlay test",
        description="",
        type="task",
        status="open",
        priority=2,
        assignee=None,
        creator=None,
        parent=None,
        labels=[],
        dependencies=[],
        comments=[],
        created_at=updated_at,
        updated_at=updated_at,
        closed_at=None,
        custom={},
    )


def test_overlay_prefers_newer_overlay() -> None:
    base_time = datetime.now(timezone.utc) - timedelta(minutes=30)
    overlay_time = base_time + timedelta(hours=1)
    base_issue = _issue("kanbus-1", base_time)
    overlay_issue = _issue("kanbus-1", overlay_time)
    overlay_record = OverlayIssueRecord(
        issue=overlay_issue,
        overlay_ts=overlay_time.isoformat().replace("+00:00", "Z"),
        overlay_event_id=None,
    )
    with tempfile.TemporaryDirectory() as tmp:
        project_dir = Path(tmp) / "project"
        project_dir.mkdir(parents=True, exist_ok=True)
        result = resolve_issue_with_overlay(
            project_dir,
            base_issue,
            overlay_record,
            None,
            OverlayConfig(enabled=True, ttl_s=86400),
        )
    assert result is not None
    assert result.identifier == overlay_issue.identifier
    assert result.updated_at == overlay_issue.updated_at


def test_gc_overlay_removes_stale_overlay() -> None:
    base_time = datetime.now(timezone.utc)
    overlay_time = base_time - timedelta(hours=1)
    base_issue = _issue("kanbus-2", base_time)
    overlay_issue = _issue("kanbus-2", overlay_time)
    with tempfile.TemporaryDirectory() as tmp:
        project_dir = Path(tmp) / "project"
        issues_dir = project_dir / "issues"
        issues_dir.mkdir(parents=True, exist_ok=True)
        issue_path = issues_dir / "kanbus-2.json"
        issue_path.write_text(
            json.dumps(base_issue.model_dump(by_alias=True, mode="json"), indent=2),
            encoding="utf-8",
        )
        write_overlay_issue(
            project_dir,
            overlay_issue,
            overlay_time.isoformat().replace("+00:00", "Z"),
            None,
        )
        gc_overlay(project_dir, OverlayConfig(enabled=True, ttl_s=86400))
        assert not overlay_issue_path(project_dir, "kanbus-2").exists()


def test_reconcile_prunes_and_removes_resolved_overlay() -> None:
    base_time = datetime.now(timezone.utc)
    with tempfile.TemporaryDirectory() as tmp:
        project_dir = Path(tmp) / "project"
        issues_dir = project_dir / "issues"
        issues_dir.mkdir(parents=True, exist_ok=True)
        base_issue = _issue("kanbus-3", base_time)
        issue_path = issues_dir / "kanbus-3.json"
        issue_path.write_text(
            json.dumps(base_issue.model_dump(by_alias=True, mode="json"), indent=2),
            encoding="utf-8",
        )
        overlay_path = overlay_issue_path(project_dir, "kanbus-3")
        overlay_path.parent.mkdir(parents=True, exist_ok=True)
        payload = {
            "overlay_ts": base_time.isoformat().replace("+00:00", "Z"),
            "overlay_event_id": "evt-1",
            "overrides": {"title": "Overlay test"},
            "issue": base_issue.model_dump(by_alias=True, mode="json"),
        }
        overlay_path.write_text(json.dumps(payload, indent=2), encoding="utf-8")
        stats = reconcile_overlay(
            project_dir,
            OverlayConfig(enabled=True, ttl_s=86400),
            prune=True,
            dry_run=False,
        )
        assert stats.issues_scanned == 1
        assert stats.issues_removed == 1
        assert stats.fields_pruned == 1
        assert not overlay_path.exists()


def test_install_overlay_hooks_creates_and_is_idempotent() -> None:
    with tempfile.TemporaryDirectory() as tmp:
        root = Path(tmp)
        subprocess.run(["git", "init"], cwd=root, check=True, capture_output=True)

        install_overlay_hooks(root)
        install_overlay_hooks(root)

        hook_path = root / ".git" / "hooks" / "post-merge"
        contents = hook_path.read_text(encoding="utf-8")
        assert "# Kanbus overlay cache maintenance" in contents
        assert contents.count("# Kanbus overlay cache maintenance") == 1
        assert hook_path.stat().st_mode & 0o111 != 0


def test_install_overlay_hooks_fails_when_hooks_disabled() -> None:
    with tempfile.TemporaryDirectory() as tmp:
        root = Path(tmp)
        subprocess.run(["git", "init"], cwd=root, check=True, capture_output=True)
        subprocess.run(
            ["git", "config", "core.hooksPath", "/dev/null"],
            cwd=root,
            check=True,
            capture_output=True,
        )

        try:
            install_overlay_hooks(root)
        except RuntimeError as error:
            assert "core.hooksPath=/dev/null" in str(error)
        else:
            raise AssertionError("expected RuntimeError")


def test_install_overlay_hooks_fails_when_hooks_path_is_file() -> None:
    with tempfile.TemporaryDirectory() as tmp:
        root = Path(tmp)
        subprocess.run(["git", "init"], cwd=root, check=True, capture_output=True)
        hooks_file = root / "hooks-file"
        hooks_file.write_text("not a directory\n", encoding="utf-8")
        subprocess.run(
            ["git", "config", "core.hooksPath", "hooks-file"],
            cwd=root,
            check=True,
            capture_output=True,
        )

        try:
            install_overlay_hooks(root)
        except RuntimeError as error:
            assert "not a directory" in str(error)
        else:
            raise AssertionError("expected RuntimeError")


def test_overlay_time_helpers_handle_invalid_and_equal_timestamps() -> None:
    now = datetime.now(timezone.utc)

    assert overlay_module._parse_ts("not-a-timestamp") is None
    assert not overlay_module._is_expired("not-a-timestamp", 10, now)
    assert overlay_module._overlay_is_newer(
        now.isoformat().replace("+00:00", "Z"),
        now,
        "evt-2",
        "evt-1",
    )


def test_overlay_time_helpers_cover_tombstone_base_comparisons() -> None:
    now = datetime.now(timezone.utc)
    stale_ts = (now - timedelta(minutes=2)).isoformat().replace("+00:00", "Z")
    future_ts = (now + timedelta(minutes=2)).isoformat().replace("+00:00", "Z")

    assert overlay_module._tombstone_newer_than_base(future_ts, now) is True
    assert overlay_module._tombstone_newer_than_base(stale_ts, now) is False
    assert overlay_module._tombstone_newer_than_base(stale_ts, None) is True
    assert overlay_module._base_newer_than_tombstone(now, stale_ts) is True
    assert overlay_module._base_newer_than_tombstone(now, "bad-ts") is True


def test_overlay_hook_block_contains_kanbus_and_kbs_commands() -> None:
    block = overlay_module._overlay_hook_block()
    assert "kanbus overlay reconcile --all --prune" in block
    assert "kbs overlay gc --all" in block


def test_apply_overrides_returns_none_for_invalid_issue_payload() -> None:
    issue = _issue("kanbus-9", datetime.now(timezone.utc))
    # Priority must be an integer, so this payload cannot validate to IssueData.
    result = overlay_module._apply_overrides(issue, {"priority": "invalid"})
    assert result is None


def test_project_reconcile_gc_reject_conflicting_selector_flags() -> None:
    with tempfile.TemporaryDirectory() as tmp:
        root = Path(tmp)
        try:
            overlay_module.gc_overlay_for_projects(
                root, project_label="alpha", all_projects=True
            )
        except ValueError as error:
            assert "cannot combine --project with --all" in str(error)
        else:
            raise AssertionError("expected ValueError")


def test_gc_overlay_for_projects_selects_default_label(
    monkeypatch, tmp_path: Path
) -> None:
    configuration = SimpleNamespace(
        project_key="alpha",
        overlay=OverlayConfig(enabled=True, ttl_s=120),
    )
    labeled = [
        SimpleNamespace(label="alpha", project_dir=tmp_path / "alpha"),
        SimpleNamespace(label="beta", project_dir=tmp_path / "beta"),
    ]
    monkeypatch.setattr(
        overlay_module, "get_configuration_path", lambda _root: tmp_path / ".kanbus.yml"
    )
    monkeypatch.setattr(
        overlay_module, "load_project_configuration", lambda _path: configuration
    )
    monkeypatch.setattr(
        overlay_module, "resolve_labeled_projects", lambda _root: labeled
    )
    seen: list[Path] = []
    monkeypatch.setattr(
        overlay_module,
        "gc_overlay",
        lambda project_dir, _config: seen.append(project_dir),
    )
    count = overlay_module.gc_overlay_for_projects(
        tmp_path, project_label=None, all_projects=False
    )
    assert count == 1
    assert seen == [tmp_path / "alpha"]


def test_gc_overlay_for_projects_returns_zero_when_no_projects(
    monkeypatch, tmp_path: Path
) -> None:
    configuration = SimpleNamespace(
        project_key="alpha",
        overlay=OverlayConfig(enabled=True, ttl_s=120),
    )
    monkeypatch.setattr(
        overlay_module, "get_configuration_path", lambda _root: tmp_path / ".kanbus.yml"
    )
    monkeypatch.setattr(
        overlay_module, "load_project_configuration", lambda _path: configuration
    )
    monkeypatch.setattr(overlay_module, "resolve_labeled_projects", lambda _root: [])
    assert (
        overlay_module.gc_overlay_for_projects(
            tmp_path, project_label=None, all_projects=False
        )
        == 0
    )


def test_reconcile_overlay_for_projects_aggregates_stats(
    monkeypatch, tmp_path: Path
) -> None:
    configuration = SimpleNamespace(
        project_key="alpha",
        overlay=OverlayConfig(enabled=True, ttl_s=120),
    )
    labeled = [
        SimpleNamespace(label="alpha", project_dir=tmp_path / "alpha"),
        SimpleNamespace(label="beta", project_dir=tmp_path / "beta"),
    ]
    monkeypatch.setattr(
        overlay_module, "get_configuration_path", lambda _root: tmp_path / ".kanbus.yml"
    )
    monkeypatch.setattr(
        overlay_module, "load_project_configuration", lambda _path: configuration
    )
    monkeypatch.setattr(
        overlay_module, "resolve_labeled_projects", lambda _root: labeled
    )
    monkeypatch.setattr(
        overlay_module,
        "reconcile_overlay",
        lambda project_dir, config, prune, dry_run: overlay_module.OverlayReconcileStats(
            issues_scanned=2 if project_dir.name == "alpha" else 3,
            issues_updated=1,
            issues_removed=0,
            fields_pruned=4,
        ),
    )
    stats = overlay_module.reconcile_overlay_for_projects(
        tmp_path,
        project_label=None,
        all_projects=True,
        prune=True,
        dry_run=False,
    )
    assert stats.projects == 2
    assert stats.issues_scanned == 5
    assert stats.issues_updated == 2
    assert stats.fields_pruned == 8


def test_reconcile_overlay_for_projects_rejects_unknown_label(
    monkeypatch, tmp_path: Path
) -> None:
    configuration = SimpleNamespace(
        project_key="alpha",
        overlay=OverlayConfig(enabled=True, ttl_s=120),
    )
    monkeypatch.setattr(
        overlay_module, "get_configuration_path", lambda _root: tmp_path / ".kanbus.yml"
    )
    monkeypatch.setattr(
        overlay_module, "load_project_configuration", lambda _path: configuration
    )
    monkeypatch.setattr(
        overlay_module,
        "resolve_labeled_projects",
        lambda _root: [SimpleNamespace(label="alpha", project_dir=tmp_path / "alpha")],
    )
    try:
        overlay_module.reconcile_overlay_for_projects(
            tmp_path,
            project_label="missing",
            all_projects=False,
            prune=False,
            dry_run=True,
        )
    except ValueError as error:
        assert "unknown project label" in str(error)
    else:
        raise AssertionError("expected ValueError")


def test_resolve_git_hooks_dir_handles_non_repo_and_relative_path(
    monkeypatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(
        overlay_module.subprocess,
        "run",
        lambda *args, **kwargs: SimpleNamespace(returncode=1, stdout="", stderr="nope"),
    )
    try:
        overlay_module._resolve_git_hooks_dir(tmp_path)
    except RuntimeError as error:
        assert "not a git repository" in str(error)
    else:
        raise AssertionError("expected RuntimeError")

    monkeypatch.setattr(
        overlay_module.subprocess,
        "run",
        lambda *args, **kwargs: SimpleNamespace(returncode=0, stdout=".git/hooks\n"),
    )
    assert (
        overlay_module._resolve_git_hooks_dir(tmp_path) == tmp_path / ".git" / "hooks"
    )


def test_issue_map_helpers_and_tagging_cover_edges() -> None:
    now = datetime.now(timezone.utc)
    issue = _issue("kanbus-helpers", now)
    issue.custom["source"] = "local"
    issue.custom["last_event_id"] = 123

    tagged = overlay_module._tag_issue(issue, "alpha")
    assert tagged is not None
    assert tagged.custom["source"] == "local"
    assert tagged.custom["project_label"] == "alpha"
    assert overlay_module._extract_event_id(issue) is None

    payload = overlay_module._issue_to_map(issue)
    assert overlay_module._issue_from_map(payload) is not None
    assert overlay_module._issue_from_map({"id": "broken"}) is None


def test_diff_issue_fields_and_remove_path_and_ensure_executable(
    monkeypatch, tmp_path: Path
) -> None:
    now = datetime.now(timezone.utc)
    base = _issue("kanbus-a", now)
    overlay = base.model_copy(update={"title": "Updated title", "priority": 1})
    diff = overlay_module._diff_issue_fields(base, overlay)
    assert diff["title"] == "Updated title"
    assert diff["priority"] == 1

    missing = tmp_path / "missing.json"
    overlay_module._remove_path(missing)
    assert not missing.exists()

    marker = tmp_path / "hook.sh"
    marker.write_text("#!/bin/sh\necho test\n", encoding="utf-8")
    overlay_module._ensure_executable(marker)
    assert marker.stat().st_mode & 0o111 != 0

    monkeypatch.setattr(
        Path,
        "chmod",
        lambda self, _mode: (_ for _ in ()).throw(OSError("denied")),
    )
    overlay_module._ensure_executable(marker)


def test_overlay_path_and_tombstone_round_trip(tmp_path: Path) -> None:
    project_dir = tmp_path / "project"
    issue = _issue("kanbus-roundtrip", datetime.now(timezone.utc))
    overlay_module.write_overlay_issue(
        project_dir,
        issue,
        "2026-01-01T00:00:00Z",
        "evt-1",
    )
    tombstone = overlay_module.OverlayTombstone(
        op="delete",
        project="alpha",
        id="kanbus-roundtrip",
        event_id="evt-2",
        ts="2026-01-01T00:00:01Z",
        ttl_s=60,
    )
    overlay_module.write_tombstone(project_dir, tombstone)
    loaded_overlay = overlay_module.load_overlay_issue(project_dir, "kanbus-roundtrip")
    loaded_tombstone = overlay_module.load_tombstone(project_dir, "kanbus-roundtrip")
    assert loaded_overlay is not None
    assert loaded_overlay.overlay_event_id == "evt-1"
    assert loaded_tombstone is not None
    assert loaded_tombstone.project == "alpha"
    assert overlay_module.overlay_tombstone_path(
        project_dir, "kanbus-roundtrip"
    ).exists()


def test_resolve_issue_with_overlay_handles_disabled_and_tombstone_paths(
    tmp_path: Path,
) -> None:
    project_dir = tmp_path / "project"
    now = datetime.now(timezone.utc)
    base = _issue("kanbus-resolve", now)
    assert (
        overlay_module.resolve_issue_with_overlay(
            project_dir,
            base,
            None,
            None,
            OverlayConfig(enabled=False, ttl_s=60),
        )
        == base
    )
    tombstone = overlay_module.OverlayTombstone(
        op="delete",
        project="alpha",
        id="kanbus-resolve",
        event_id="evt",
        ts=(now + timedelta(minutes=1)).isoformat().replace("+00:00", "Z"),
        ttl_s=60,
    )
    assert (
        overlay_module.resolve_issue_with_overlay(
            project_dir,
            base,
            None,
            tombstone,
            OverlayConfig(enabled=True, ttl_s=60),
        )
        is None
    )


def test_apply_overlay_to_issues_handles_disabled_local_and_overlay_only_entries(
    tmp_path: Path,
) -> None:
    project_dir = tmp_path / "project"
    project_dir.mkdir(parents=True, exist_ok=True)
    now = datetime.now(timezone.utc)
    shared = _issue("kanbus-shared", now)
    local = _issue("kanbus-local", now)
    local.custom["source"] = "local"

    disabled = overlay_module.apply_overlay_to_issues(
        project_dir, [shared, local], OverlayConfig(enabled=False, ttl_s=60)
    )
    assert [item.identifier for item in disabled] == ["kanbus-shared", "kanbus-local"]

    overlay_only = _issue("kanbus-only", now + timedelta(minutes=2))
    overlay_module.write_overlay_issue(
        project_dir,
        overlay_only,
        overlay_only.updated_at.isoformat().replace("+00:00", "Z"),
        "evt-only",
    )
    applied = overlay_module.apply_overlay_to_issues(
        project_dir,
        [shared, local],
        OverlayConfig(enabled=True, ttl_s=3600),
        project_label="alpha",
    )
    identifiers = [item.identifier for item in applied]
    assert "kanbus-local" in identifiers
    assert "kanbus-only" in identifiers
    assert identifiers == sorted(identifiers)


def test_gc_overlay_handles_disabled_and_invalid_overlay_entries(
    monkeypatch, tmp_path: Path
) -> None:
    project_dir = tmp_path / "project"
    overlay_issue_dir = project_dir / ".overlay" / "issues"
    overlay_tombstone_dir = project_dir / ".overlay" / "tombstones"
    overlay_issue_dir.mkdir(parents=True, exist_ok=True)
    overlay_tombstone_dir.mkdir(parents=True, exist_ok=True)
    stale_overlay = overlay_issue_dir / "stale.json"
    stale_overlay.write_text("{}", encoding="utf-8")
    stale_tombstone = overlay_tombstone_dir / "stale.json"
    stale_tombstone.write_text("{}", encoding="utf-8")

    overlay_module.gc_overlay(project_dir, OverlayConfig(enabled=False, ttl_s=60))
    assert stale_overlay.exists()

    monkeypatch.setattr(
        overlay_module,
        "load_overlay_issue",
        lambda _project_dir, _issue_id: None,
    )
    monkeypatch.setattr(
        overlay_module,
        "load_tombstone",
        lambda _project_dir, _issue_id: None,
    )
    overlay_module.gc_overlay(project_dir, OverlayConfig(enabled=True, ttl_s=60))
    assert not stale_overlay.exists()
    assert not stale_tombstone.exists()


def test_resolve_issue_with_overlay_covers_expired_and_override_merge_paths(
    tmp_path: Path,
) -> None:
    project_dir = tmp_path / "project"
    now = datetime.now(timezone.utc)
    base = _issue("kanbus-merge", now)
    base.custom["last_event_id"] = "evt-1"
    overlay_issue = _issue("kanbus-merge", now + timedelta(minutes=5))
    overlay_record = overlay_module.OverlayIssueRecord(
        issue=overlay_issue,
        overlay_ts=(now + timedelta(minutes=5)).isoformat().replace("+00:00", "Z"),
        overlay_event_id="evt-2",
        overrides={"title": "Merged title"},
    )
    merged = overlay_module.resolve_issue_with_overlay(
        project_dir,
        base,
        overlay_record,
        None,
        OverlayConfig(enabled=True, ttl_s=3600),
        project_label="alpha",
    )
    assert merged is not None
    assert merged.title == "Merged title"
    assert merged.custom["project_label"] == "alpha"

    expired = overlay_module.OverlayIssueRecord(
        issue=overlay_issue,
        overlay_ts=(now - timedelta(days=2)).isoformat().replace("+00:00", "Z"),
        overlay_event_id="evt-old",
    )
    resolved = overlay_module.resolve_issue_with_overlay(
        project_dir,
        base,
        expired,
        None,
        OverlayConfig(enabled=True, ttl_s=1),
    )
    assert resolved is not None
    assert resolved.identifier == "kanbus-merge"


def test_apply_overlay_to_issues_skips_base_collision_and_missing_overlay(
    tmp_path: Path, monkeypatch
) -> None:
    project_dir = tmp_path / "project"
    overlay_dir = project_dir / ".overlay" / "issues"
    overlay_dir.mkdir(parents=True, exist_ok=True)
    now = datetime.now(timezone.utc)
    shared = _issue("kanbus-shared2", now)
    overlay_module.write_overlay_issue(
        project_dir,
        _issue("kanbus-shared2", now + timedelta(minutes=1)),
        (now + timedelta(minutes=1)).isoformat().replace("+00:00", "Z"),
        "evt-a",
    )
    overlay_module.write_overlay_issue(
        project_dir,
        _issue("kanbus-missing-overlay", now + timedelta(minutes=2)),
        (now + timedelta(minutes=2)).isoformat().replace("+00:00", "Z"),
        "evt-b",
    )
    real_load = overlay_module.load_overlay_issue
    monkeypatch.setattr(
        overlay_module,
        "load_overlay_issue",
        lambda p, issue_id: (
            None if issue_id == "kanbus-missing-overlay" else real_load(p, issue_id)
        ),
    )
    results = overlay_module.apply_overlay_to_issues(
        project_dir,
        [shared],
        OverlayConfig(enabled=True, ttl_s=3600),
    )
    assert [item.identifier for item in results] == ["kanbus-shared2"]


def test_gc_overlay_removes_expired_and_base_newer_records(tmp_path: Path) -> None:
    project_dir = tmp_path / "project"
    base_issues = project_dir / "issues"
    base_issues.mkdir(parents=True, exist_ok=True)
    now = datetime.now(timezone.utc)

    expired_issue = _issue("kanbus-expired", now - timedelta(days=2))
    overlay_module.write_overlay_issue(
        project_dir,
        expired_issue,
        expired_issue.updated_at.isoformat().replace("+00:00", "Z"),
        "evt-expired",
    )

    base_newer = _issue("kanbus-base-newer", now)
    (base_issues / "kanbus-base-newer.json").write_text(
        json.dumps(base_newer.model_dump(by_alias=True, mode="json"), indent=2),
        encoding="utf-8",
    )
    old_overlay = _issue("kanbus-base-newer", now - timedelta(minutes=20))
    overlay_module.write_overlay_issue(
        project_dir,
        old_overlay,
        old_overlay.updated_at.isoformat().replace("+00:00", "Z"),
        "evt-old",
    )

    tombstone_expired = overlay_module.OverlayTombstone(
        op="delete",
        project="alpha",
        id="kanbus-tomb-expired",
        event_id="evt-t1",
        ts=(now - timedelta(days=2)).isoformat().replace("+00:00", "Z"),
        ttl_s=1,
    )
    overlay_module.write_tombstone(project_dir, tombstone_expired)

    tombstone_old = overlay_module.OverlayTombstone(
        op="delete",
        project="alpha",
        id="kanbus-base-newer",
        event_id="evt-t2",
        ts=(now - timedelta(hours=2)).isoformat().replace("+00:00", "Z"),
        ttl_s=3600,
    )
    overlay_module.write_tombstone(project_dir, tombstone_old)

    overlay_module.gc_overlay(project_dir, OverlayConfig(enabled=True, ttl_s=60))
    assert not overlay_module.overlay_issue_path(project_dir, "kanbus-expired").exists()
    assert not overlay_module.overlay_issue_path(
        project_dir, "kanbus-base-newer"
    ).exists()
    assert not overlay_module.overlay_tombstone_path(
        project_dir, "kanbus-tomb-expired"
    ).exists()
    assert not overlay_module.overlay_tombstone_path(
        project_dir, "kanbus-base-newer"
    ).exists()


def test_gc_overlay_tolerates_invalid_base_issue_json(tmp_path: Path) -> None:
    project_dir = tmp_path / "project"
    base_issues = project_dir / "issues"
    base_issues.mkdir(parents=True, exist_ok=True)
    now = datetime.now(timezone.utc)
    overlay_module.write_overlay_issue(
        project_dir,
        _issue("kanbus-invalid-base", now),
        now.isoformat().replace("+00:00", "Z"),
        "evt",
    )
    (base_issues / "kanbus-invalid-base.json").write_text("{", encoding="utf-8")
    overlay_module.gc_overlay(project_dir, OverlayConfig(enabled=True, ttl_s=3600))
    assert overlay_module.overlay_issue_path(
        project_dir, "kanbus-invalid-base"
    ).exists()


def test_reconcile_overlay_returns_defaults_for_disabled_or_missing_dir(
    tmp_path: Path,
) -> None:
    project_dir = tmp_path / "project"
    disabled = overlay_module.reconcile_overlay(
        project_dir, OverlayConfig(enabled=False, ttl_s=60), prune=False, dry_run=False
    )
    assert disabled == overlay_module.OverlayReconcileStats()

    enabled_no_dir = overlay_module.reconcile_overlay(
        project_dir, OverlayConfig(enabled=True, ttl_s=60), prune=False, dry_run=False
    )
    assert enabled_no_dir == overlay_module.OverlayReconcileStats()


def test_reconcile_overlay_covers_remove_update_and_dry_run_paths(
    tmp_path: Path, monkeypatch
) -> None:
    project_dir = tmp_path / "project"
    base_issues = project_dir / "issues"
    base_issues.mkdir(parents=True, exist_ok=True)
    overlay_dir = project_dir / ".overlay" / "issues"
    overlay_dir.mkdir(parents=True, exist_ok=True)
    now = datetime.now(timezone.utc)

    base_same = _issue("kanbus-same", now)
    (base_issues / "kanbus-same.json").write_text(
        json.dumps(base_same.model_dump(by_alias=True, mode="json"), indent=2),
        encoding="utf-8",
    )
    overlay_module.write_overlay_issue(
        project_dir,
        base_same,
        now.isoformat().replace("+00:00", "Z"),
        "evt-same",
    )

    base_update = _issue("kanbus-update", now)
    (base_issues / "kanbus-update.json").write_text(
        json.dumps(base_update.model_dump(by_alias=True, mode="json"), indent=2),
        encoding="utf-8",
    )
    changed = base_update.model_copy(update={"title": "Changed"})
    overlay_path = overlay_module.overlay_issue_path(project_dir, "kanbus-update")
    overlay_path.parent.mkdir(parents=True, exist_ok=True)
    overlay_path.write_text(
        json.dumps(
            {
                "overlay_ts": now.isoformat().replace("+00:00", "Z"),
                "overlay_event_id": "evt-update",
                "overrides": None,
                "issue": changed.model_dump(by_alias=True, mode="json"),
            },
            indent=2,
        ),
        encoding="utf-8",
    )

    orphan_issue = _issue("orphan", now)
    overlay_module.write_overlay_issue(
        project_dir,
        orphan_issue,
        now.isoformat().replace("+00:00", "Z"),
        "evt-orphan",
    )
    missing_base_overlay = overlay_dir / "missing-base.json"
    missing_base_overlay.write_text(
        json.dumps(
            {
                "overlay_ts": now.isoformat().replace("+00:00", "Z"),
                "overlay_event_id": "evt-missing",
                "overrides": {"title": "X"},
                "issue": base_update.model_dump(by_alias=True, mode="json"),
            },
            indent=2,
        ),
        encoding="utf-8",
    )
    invalid_base_overlay = overlay_dir / "invalid-base.json"
    invalid_base_overlay.write_text(
        json.dumps(
            {
                "overlay_ts": now.isoformat().replace("+00:00", "Z"),
                "overlay_event_id": "evt-invalid",
                "overrides": {"title": "Y"},
                "issue": base_update.model_dump(by_alias=True, mode="json"),
            },
            indent=2,
        ),
        encoding="utf-8",
    )
    (base_issues / "invalid-base.json").write_text("{", encoding="utf-8")

    real_load = overlay_module.load_overlay_issue
    # Force one path through the overlay_issue is None branch.
    monkeypatch.setattr(
        overlay_module,
        "load_overlay_issue",
        lambda p, issue_id: None if issue_id == "orphan" else real_load(p, issue_id),
    )
    stats = overlay_module.reconcile_overlay(
        project_dir, OverlayConfig(enabled=True, ttl_s=60), prune=True, dry_run=False
    )
    assert stats.issues_scanned >= 2
    assert stats.issues_updated >= 1
    assert stats.issues_removed >= 1
    assert not overlay_module.overlay_issue_path(project_dir, "orphan").exists()
    assert not overlay_module.overlay_issue_path(project_dir, "kanbus-same").exists()
    updated_payload = json.loads(overlay_path.read_text(encoding="utf-8"))
    assert updated_payload["overrides"]["title"] == "Changed"

    dry_run_overlay = overlay_module.overlay_issue_path(project_dir, "kanbus-dry")
    base_dry = _issue("kanbus-dry", now)
    (base_issues / "kanbus-dry.json").write_text(
        json.dumps(base_dry.model_dump(by_alias=True, mode="json"), indent=2),
        encoding="utf-8",
    )
    dry_run_overlay.parent.mkdir(parents=True, exist_ok=True)
    dry_overlay_issue = base_dry.model_copy(update={"title": "Dry Run Changed"})
    dry_run_overlay.write_text(
        json.dumps(
            {
                "overlay_ts": now.isoformat().replace("+00:00", "Z"),
                "overlay_event_id": "evt-dry",
                "overrides": None,
                "issue": dry_overlay_issue.model_dump(by_alias=True, mode="json"),
            },
            indent=2,
        ),
        encoding="utf-8",
    )
    stats_dry = overlay_module.reconcile_overlay(
        project_dir, OverlayConfig(enabled=True, ttl_s=60), prune=False, dry_run=True
    )
    assert stats_dry.issues_removed >= 0
    assert dry_run_overlay.exists()


def test_overlay_remaining_helper_branches(monkeypatch, tmp_path: Path) -> None:
    now = datetime.now(timezone.utc)
    assert overlay_module._tag_issue(None, "alpha") is None
    assert overlay_module._parse_ts("") is None
    assert overlay_module._is_expired("2026-01-01T00:00:00Z", 0, now) is False
    assert overlay_module._overlay_is_newer("bad-ts", now, None, None) is False
    assert (
        overlay_module._overlay_is_newer(
            now.isoformat().replace("+00:00", "Z"), now, None, None
        )
        is True
    )
    assert overlay_module._tombstone_newer_than_base("bad-ts", now) is False

    issue = _issue("kanbus-evt", now)
    issue.custom["last_event_id"] = "evt-1"
    assert overlay_module._extract_event_id(issue) == "evt-1"

    marker = tmp_path / "marker.txt"
    marker.write_text("x", encoding="utf-8")
    monkeypatch.setattr(
        Path,
        "unlink",
        lambda self, missing_ok=True: (_ for _ in ()).throw(OSError("denied")),
    )
    overlay_module._remove_path(marker)


def test_resolve_issue_with_overlay_covers_remove_paths(tmp_path: Path) -> None:
    project_dir = tmp_path / "project"
    now = datetime.now(timezone.utc)
    base = _issue("kanbus-paths", now)
    overlay_issue = _issue("kanbus-paths", now - timedelta(minutes=1))
    overlay_record = overlay_module.OverlayIssueRecord(
        issue=overlay_issue,
        overlay_ts=overlay_issue.updated_at.isoformat().replace("+00:00", "Z"),
        overlay_event_id="evt-old",
    )
    overlay_module.write_overlay_issue(
        project_dir,
        overlay_issue,
        overlay_record.overlay_ts,
        overlay_record.overlay_event_id,
    )
    result = overlay_module.resolve_issue_with_overlay(
        project_dir,
        base,
        overlay_record,
        None,
        OverlayConfig(enabled=True, ttl_s=3600),
    )
    assert result is not None
    assert result.identifier == "kanbus-paths"
    assert not overlay_module.overlay_issue_path(project_dir, "kanbus-paths").exists()

    tombstone = overlay_module.OverlayTombstone(
        op="delete",
        project="alpha",
        id="kanbus-paths",
        event_id="evt-tomb",
        ts=(now - timedelta(days=2)).isoformat().replace("+00:00", "Z"),
        ttl_s=1,
    )
    overlay_module.write_tombstone(project_dir, tombstone)
    resolved = overlay_module.resolve_issue_with_overlay(
        project_dir,
        base,
        None,
        tombstone,
        OverlayConfig(enabled=True, ttl_s=3600),
    )
    assert resolved is not None
    assert not overlay_module.overlay_tombstone_path(
        project_dir, "kanbus-paths"
    ).exists()


def test_apply_overlay_to_issues_includes_overlay_only_resolved_issue(
    tmp_path: Path,
) -> None:
    project_dir = tmp_path / "project"
    now = datetime.now(timezone.utc)
    shared = _issue("kanbus-base", now)
    overlay_only = _issue("kanbus-overlay-only", now + timedelta(minutes=3))
    overlay_module.write_overlay_issue(
        project_dir,
        overlay_only,
        overlay_only.updated_at.isoformat().replace("+00:00", "Z"),
        "evt-overlay-only",
    )
    result = overlay_module.apply_overlay_to_issues(
        project_dir,
        [shared],
        OverlayConfig(enabled=True, ttl_s=3600),
        project_label="alpha",
    )
    ids = [item.identifier for item in result]
    assert "kanbus-overlay-only" in ids


def test_gc_overlay_tombstone_invalid_base_and_base_newer(tmp_path: Path) -> None:
    project_dir = tmp_path / "project"
    base_issues = project_dir / "issues"
    base_issues.mkdir(parents=True, exist_ok=True)
    now = datetime.now(timezone.utc)

    bad_base_tombstone = overlay_module.OverlayTombstone(
        op="delete",
        project="alpha",
        id="kanbus-bad-base",
        event_id="evt-bad",
        ts=(now - timedelta(minutes=10)).isoformat().replace("+00:00", "Z"),
        ttl_s=3600,
    )
    overlay_module.write_tombstone(project_dir, bad_base_tombstone)
    (base_issues / "kanbus-bad-base.json").write_text("{", encoding="utf-8")

    base_new = _issue("kanbus-remove-tomb", now)
    (base_issues / "kanbus-remove-tomb.json").write_text(
        json.dumps(base_new.model_dump(by_alias=True, mode="json"), indent=2),
        encoding="utf-8",
    )
    removable = overlay_module.OverlayTombstone(
        op="delete",
        project="alpha",
        id="kanbus-remove-tomb",
        event_id="evt-remove",
        ts=(now - timedelta(hours=1)).isoformat().replace("+00:00", "Z"),
        ttl_s=999999,
    )
    overlay_module.write_tombstone(project_dir, removable)
    overlay_module.gc_overlay(project_dir, OverlayConfig(enabled=True, ttl_s=3600))
    assert overlay_module.overlay_tombstone_path(
        project_dir, "kanbus-bad-base"
    ).exists()
    assert not overlay_module.overlay_tombstone_path(
        project_dir, "kanbus-remove-tomb"
    ).exists()


def test_reconcile_overlay_covers_merge_none_and_selector_branches(
    monkeypatch, tmp_path: Path
) -> None:
    project_dir = tmp_path / "project"
    base_issues = project_dir / "issues"
    base_issues.mkdir(parents=True, exist_ok=True)
    now = datetime.now(timezone.utc)
    base = _issue("kanbus-merge-none", now)
    (base_issues / "kanbus-merge-none.json").write_text(
        json.dumps(base.model_dump(by_alias=True, mode="json"), indent=2),
        encoding="utf-8",
    )
    overlay_module.write_overlay_issue(
        project_dir,
        base.model_copy(update={"title": "Other"}),
        now.isoformat().replace("+00:00", "Z"),
        "evt",
    )
    monkeypatch.setattr(
        overlay_module, "_apply_overrides", lambda *_args, **_kwargs: None
    )
    stats = overlay_module.reconcile_overlay(
        project_dir, OverlayConfig(enabled=True, ttl_s=60), prune=False, dry_run=False
    )
    assert stats.issues_updated == 0

    configuration = SimpleNamespace(
        project_key="alpha",
        overlay=OverlayConfig(enabled=True, ttl_s=60),
    )
    labeled = [SimpleNamespace(label="alpha", project_dir=project_dir)]
    monkeypatch.setattr(
        overlay_module, "get_configuration_path", lambda _root: tmp_path / ".kanbus.yml"
    )
    monkeypatch.setattr(
        overlay_module, "load_project_configuration", lambda _path: configuration
    )
    monkeypatch.setattr(
        overlay_module, "resolve_labeled_projects", lambda _root: labeled
    )
    assert overlay_module.gc_overlay_for_projects(tmp_path, None, True) == 1
    try:
        overlay_module.gc_overlay_for_projects(tmp_path, "missing", False)
    except ValueError as error:
        assert "unknown project label" in str(error)
    else:
        raise AssertionError("expected ValueError")

    monkeypatch.setattr(overlay_module, "resolve_labeled_projects", lambda _root: [])
    assert (
        overlay_module.reconcile_overlay_for_projects(
            tmp_path, None, False, prune=False, dry_run=False
        )
        == overlay_module.OverlayReconcileStats()
    )
    try:
        overlay_module.reconcile_overlay_for_projects(
            tmp_path, "alpha", True, prune=False, dry_run=False
        )
    except ValueError as error:
        assert "cannot combine --project with --all" in str(error)
    else:
        raise AssertionError("expected ValueError")


def test_install_overlay_hooks_appends_block_when_existing_hook_has_other_content(
    tmp_path: Path, monkeypatch
) -> None:
    root = tmp_path
    hooks_dir = root / ".git" / "hooks"
    hooks_dir.mkdir(parents=True, exist_ok=True)
    hook = hooks_dir / "post-merge"
    hook.write_text("#!/bin/sh\necho hello\n", encoding="utf-8")
    monkeypatch.setattr(
        overlay_module, "_resolve_git_hooks_dir", lambda _root: hooks_dir
    )
    overlay_module.install_overlay_hooks(root)
    contents = hook.read_text(encoding="utf-8")
    assert "echo hello" in contents
    assert "Kanbus overlay cache maintenance" in contents



"""Realtime gossip transport and envelope helpers."""

from __future__ import annotations

import json
import os
import socket
import subprocess
import sys
import time
import threading
import shutil
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Callable, Optional
from uuid import uuid4

from pydantic import BaseModel, Field, ValidationError

from kanbus.config_loader import ConfigurationError, load_project_configuration
from kanbus.models import IssueData, ProjectConfiguration, RealtimeConfig
from kanbus.overlay import OverlayTombstone, write_overlay_issue, write_tombstone
from kanbus.project import (
    ProjectMarkerError,
    get_configuration_path,
    resolve_labeled_projects,
)


class GossipEnvelope(BaseModel):
    """Realtime gossip envelope."""

    id: str = Field(min_length=1)
    ts: str = Field(min_length=1)
    project: str = Field(min_length=1)
    type: str = Field(min_length=1)
    issue_id: Optional[str] = None
    event_id: Optional[str] = None
    producer_id: str = Field(min_length=1)
    origin_cluster_id: Optional[str] = None
    issue: Optional[IssueData] = None


@dataclass(frozen=True)
class BrokerEndpoint:
    """Resolved broker endpoint."""

    scheme: str
    host: str
    port: int
    url: str


@dataclass(frozen=True)
class BrokerStartup:
    """Mosquitto startup result."""

    endpoint: BrokerEndpoint
    process: subprocess.Popen


class GossipError(RuntimeError):
    """Raised when gossip operations fail."""


_PRODUCER_ID: Optional[str] = None


def producer_id() -> str:
    """Return the producer id for this process."""
    global _PRODUCER_ID
    if _PRODUCER_ID is None:
        _PRODUCER_ID = str(uuid4())
    return _PRODUCER_ID


class DedupeSet:
    """In-memory dedupe set with TTL."""

    def __init__(self, ttl_s: int) -> None:
        self.ttl_s = ttl_s
        self._entries: dict[str, float] = {}

    def seen(self, key: str) -> bool:
        now = time.monotonic()
        self._prune(now)
        if key in self._entries:
            return True
        self._entries[key] = now
        return False

    def _prune(self, now: float) -> None:
        ttl = self.ttl_s
        expired = [key for key, ts in self._entries.items() if now - ts > ttl]
        for key in expired:
            self._entries.pop(key, None)


def publish_issue_mutation(
    root: Path,
    project_dir: Path,
    issue: IssueData,
    event_id: Optional[str],
    event_type: str,
) -> None:
    """Publish a gossip envelope for an issue mutation."""
    try:
        configuration = load_project_configuration(get_configuration_path(root))
    except (ProjectMarkerError, ConfigurationError):
        return
    if configuration.realtime.broker == "off":
        return
    project_label = _resolve_project_label(root, project_dir, configuration)
    if project_label is None:
        return
    envelope = GossipEnvelope(
        id=str(uuid4()),
        ts=_now_iso(),
        project=project_label,
        type=event_type,
        issue_id=issue.identifier,
        event_id=event_id,
        producer_id=producer_id(),
        issue=issue,
    )
    topic = configuration.realtime.topics.project_events.format(project=project_label)
    try:
        _publish_envelope(root, configuration, topic, envelope)
    except Exception as error:  # noqa: BLE001
        print(
            f"warning: realtime publish failed for {issue.identifier}: {error}",
            file=sys.stderr,
        )


def publish_issue_deleted(
    root: Path,
    project_dir: Path,
    issue_id: str,
    event_id: Optional[str],
) -> None:
    """Publish a gossip envelope for an issue deletion."""
    try:
        configuration = load_project_configuration(get_configuration_path(root))
    except (ProjectMarkerError, ConfigurationError):
        return
    if configuration.realtime.broker == "off":
        return
    project_label = _resolve_project_label(root, project_dir, configuration)
    if project_label is None:
        return
    envelope = GossipEnvelope(
        id=str(uuid4()),
        ts=_now_iso(),
        project=project_label,
        type="issue.deleted",
        issue_id=issue_id,
        event_id=event_id,
        producer_id=producer_id(),
    )
    topic = configuration.realtime.topics.project_events.format(project=project_label)
    try:
        _publish_envelope(root, configuration, topic, envelope)
    except Exception as error:  # noqa: BLE001
        print(
            f"warning: realtime publish failed for {issue_id}: {error}",
            file=sys.stderr,
        )


def run_gossip_watch(
    root: Path,
    project_filter: Optional[str],
    transport_override: Optional[str],
    broker_override: Optional[str],
    autostart_override: Optional[bool],
    keepalive_override: Optional[bool],
    print_envelopes: bool,
) -> None:
    """Subscribe to gossip notifications and update overlays."""
    _run_gossip_consumer(
        root,
        project_filter,
        transport_override,
        broker_override,
        autostart_override,
        keepalive_override,
        print_envelopes,
        on_envelope=None,
        autostart_local_uds=False,
        broker_off_is_error=True,
    )


def run_gossip_bridge(
    root: Path,
    on_envelope: Callable[[GossipEnvelope], None],
) -> None:
    """Subscribe to gossip notifications for console bridging."""
    _run_gossip_consumer(
        root,
        project_filter=None,
        transport_override=None,
        broker_override=None,
        autostart_override=None,
        keepalive_override=None,
        print_envelopes=False,
        on_envelope=on_envelope,
        autostart_local_uds=True,
        broker_off_is_error=False,
    )


def _run_gossip_consumer(
    root: Path,
    project_filter: Optional[str],
    transport_override: Optional[str],
    broker_override: Optional[str],
    autostart_override: Optional[bool],
    keepalive_override: Optional[bool],
    print_envelopes: bool,
    on_envelope: Optional[Callable[[GossipEnvelope], None]],
    autostart_local_uds: bool,
    broker_off_is_error: bool,
) -> None:
    """Shared consumer loop used by watch mode and console bridge."""
    configuration = load_project_configuration(get_configuration_path(root))
    realtime = configuration.realtime
    transport = transport_override or realtime.transport
    broker = broker_override or realtime.broker
    autostart = (
        autostart_override if autostart_override is not None else realtime.autostart
    )
    keepalive = (
        keepalive_override if keepalive_override is not None else realtime.keepalive
    )

    labeled = resolve_labeled_projects(root)
    if project_filter:
        labeled = [project for project in labeled if project.label == project_filter]
        if not labeled:
            raise GossipError(f"unknown project label: {project_filter}")

    topics = [
        realtime.topics.project_events.format(project=project.label)
        for project in labeled
    ]
    dedupe = DedupeSet(ttl_s=3600)

    def handler(envelope: GossipEnvelope) -> None:
        if dedupe.seen(envelope.id):
            return
        if envelope.producer_id == producer_id():
            return
        target_project = _resolve_project_dir(root, envelope.project)
        if target_project is None:
            return
        if print_envelopes:
            print(json.dumps(envelope.model_dump(mode="json"), sort_keys=False))
        if configuration.overlay.enabled:
            if envelope.type == "issue.mutated" and envelope.issue is not None:
                write_overlay_issue(
                    target_project,
                    envelope.issue,
                    envelope.ts,
                    envelope.event_id,
                )
            elif envelope.type == "issue.deleted" and envelope.issue_id:
                tombstone = OverlayTombstone(
                    op="delete",
                    project=envelope.project,
                    id=envelope.issue_id,
                    event_id=envelope.event_id,
                    ts=envelope.ts,
                    ttl_s=configuration.overlay.ttl_s,
                )
                write_tombstone(target_project, tombstone)
        if on_envelope is not None:
            on_envelope(envelope)

    use_uds = transport == "uds" or (
        transport == "auto" and _uds_socket_path(realtime).exists()
    )
    if autostart_local_uds and not use_uds and transport in {"auto", "uds"}:
        _ensure_local_uds_broker(realtime)
        use_uds = True
    if use_uds:
        run_uds_subscription(realtime, topics, handler)
        return
    if broker == "off":
        if broker_off_is_error:
            raise GossipError("realtime broker is disabled")
        return
    endpoint = resolve_broker_endpoint(broker)
    broker_process = None
    if not broker_is_reachable(endpoint):
        if broker == "auto":
            endpoint = _parse_broker_url("mqtt://127.0.0.1:1883")
        if not autostart:
            raise GossipError("broker not reachable and autostart disabled")
        startup = ensure_mosquitto(endpoint)
        if startup is None:
            _print_mosquitto_missing()
            return
        endpoint = startup.endpoint
        broker_process = startup.process
    run_mqtt_subscription(endpoint, topics, handler)
    if broker_process is not None and not keepalive:
        broker_process.terminate()


def _ensure_local_uds_broker(realtime: RealtimeConfig) -> None:
    socket_path = _uds_socket_path(realtime)
    if socket_path.exists():
        return
    broker_socket = socket_path
    threading.Thread(
        target=run_uds_broker,
        args=(broker_socket,),
        daemon=True,
    ).start()
    for _ in range(20):
        if socket_path.exists():
            return
        time.sleep(0.05)
    raise GossipError(f"failed to start local UDS broker at {socket_path}")


def run_gossip_broker(root: Path, socket_override: Optional[Path]) -> None:
    """Run a UDS gossip broker."""
    socket_path = socket_override
    if socket_path is None:
        try:
            configuration = load_project_configuration(get_configuration_path(root))
            socket_path = _uds_socket_path(configuration.realtime)
        except (ProjectMarkerError, ConfigurationError):
            socket_path = _uds_socket_path()
    run_uds_broker(socket_path)


def run_uds_broker(socket_path: Path) -> None:
    """Run the Unix domain socket gossip broker."""
    socket_path.parent.mkdir(parents=True, exist_ok=True)
    if socket_path.exists():
        socket_path.unlink()
    server = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    server.bind(str(socket_path))
    server.listen(64)
    subscribers: list[tuple[str, socket.socket]] = []
    lock = threading.Lock()
    while True:
        conn, _addr = server.accept()
        conn.settimeout(5.0)
        thread = threading.Thread(
            target=_handle_uds_connection, args=(conn, subscribers, lock), daemon=True
        )
        thread.start()


def _handle_uds_connection(
    conn: socket.socket,
    subscribers: list[tuple[str, socket.socket]],
    lock: threading.Lock,
) -> None:
    buffer = b""
    while True:
        try:
            chunk = conn.recv(4096)
        except socket.timeout:
            continue
        if not chunk:
            break
        buffer += chunk
        while b"\n" in buffer:
            line, buffer = buffer.split(b"\n", 1)
            if not line.strip():
                continue
            try:
                payload = json.loads(line.decode())
            except json.JSONDecodeError:
                continue
            op = payload.get("op")
            if op == "sub":
                topic = payload.get("topic")
                if isinstance(topic, str):
                    with lock:
                        subscribers.append((topic, conn))
            elif op == "pub":
                _broadcast_payload(payload, subscribers, lock)


def _broadcast_payload(
    payload: dict,
    subscribers: list[tuple[str, socket.socket]],
    lock: threading.Lock,
) -> None:
    topic = payload.get("topic")
    if not isinstance(topic, str):
        return
    message = {"topic": topic, "msg": payload.get("msg")}
    data = (json.dumps(message) + "\n").encode()
    remaining: list[tuple[str, socket.socket]] = []
    with lock:
        current = list(subscribers)
    for sub_topic, sock in current:
        if sub_topic != topic:
            remaining.append((sub_topic, sock))
            continue
        try:
            sock.sendall(data)
            remaining.append((sub_topic, sock))
        except OSError:
            continue
    with lock:
        subscribers[:] = remaining


def run_uds_subscription(
    realtime: RealtimeConfig,
    topics: list[str],
    handler: Callable[[GossipEnvelope], None],
) -> None:
    socket_path = _uds_socket_path(realtime)
    with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as sock:
        sock.connect(str(socket_path))
        for topic in topics:
            payload = json.dumps({"op": "sub", "topic": topic}) + "\n"
            sock.sendall(payload.encode())
        buffer = b""
        while True:
            chunk = sock.recv(4096)
            if not chunk:
                break
            buffer += chunk
            while b"\n" in buffer:
                line, buffer = buffer.split(b"\n", 1)
                if not line.strip():
                    continue
                try:
                    payload = json.loads(line.decode())
                    envelope = GossipEnvelope.model_validate(payload.get("msg", {}))
                except (json.JSONDecodeError, ValidationError):
                    continue
                handler(envelope)


def _publish_envelope(
    root: Path,
    configuration: ProjectConfiguration,
    topic: str,
    envelope: GossipEnvelope,
) -> None:
    transport = configuration.realtime.transport
    broker = configuration.realtime.broker
    autostart = configuration.realtime.autostart
    keepalive = configuration.realtime.keepalive
    if transport == "uds" or (
        transport == "auto" and _uds_socket_path(configuration.realtime).exists()
    ):
        try:
            _publish_uds(topic, envelope, configuration.realtime)
        except OSError:
            return
        return
    if broker == "off":
        return
    endpoint = resolve_broker_endpoint(broker)
    broker_process = None
    if not broker_is_reachable(endpoint):
        if broker == "auto":
            endpoint = _parse_broker_url("mqtt://127.0.0.1:1883")
        if not autostart:
            return
        startup = ensure_mosquitto(endpoint)
        if startup is None:
            _print_mosquitto_missing()
            return
        endpoint = startup.endpoint
        broker_process = startup.process
    _publish_mqtt(endpoint, topic, envelope)
    if broker_process is not None and not keepalive:
        broker_process.terminate()


def _publish_uds(
    topic: str, envelope: GossipEnvelope, realtime: RealtimeConfig
) -> None:
    socket_path = _uds_socket_path(realtime)
    payload = (
        json.dumps(
            {
                "op": "pub",
                "topic": topic,
                "msg": envelope.model_dump(by_alias=True, mode="json"),
            }
        )
        + "\n"
    )
    with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as sock:
        sock.settimeout(2.0)
        sock.connect(str(socket_path))
        sock.sendall(payload.encode())


def resolve_broker_endpoint(broker: str) -> BrokerEndpoint:
    if broker == "auto":
        metadata = _load_broker_metadata()
        if metadata:
            return _parse_broker_url(metadata.get("endpoint", "mqtt://127.0.0.1:1883"))
        return _parse_broker_url("mqtt://127.0.0.1:1883")
    return _parse_broker_url(broker)


def broker_is_reachable(endpoint: BrokerEndpoint) -> bool:
    try:
        with socket.create_connection((endpoint.host, endpoint.port), timeout=1.0):
            return True
    except OSError:
        return False


def ensure_mosquitto(endpoint: BrokerEndpoint) -> Optional[BrokerStartup]:
    if endpoint.scheme != "mqtt":
        return None
    if endpoint.host not in ("127.0.0.1", "localhost"):
        return None
    if not _mosquitto_available():
        return None
    run_dir = _broker_run_dir()
    run_dir.mkdir(parents=True, exist_ok=True)
    port = endpoint.port if endpoint.port > 0 else 1883
    port = _find_free_port(port)
    conf_path = run_dir / "mosquitto.conf"
    log_path = run_dir / "mosquitto.log"
    conf_path.write_text(
        "\n".join(
            [
                f"listener {port} 127.0.0.1",
                "allow_anonymous true",
                f"log_dest file {log_path}",
                "persistence false",
            ]
        )
        + "\n",
        encoding="utf-8",
    )
    process = subprocess.Popen(
        ["mosquitto", "-c", str(conf_path)],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    metadata = {
        "kind": "mosquitto",
        "endpoint": f"mqtt://127.0.0.1:{port}",
        "pid": process.pid,
        "started_by": "kanbus",
        "started_at": _now_iso(),
        "log_path": str(log_path),
        "conf_path": str(conf_path),
        "ttl_s": 86400,
    }
    _write_broker_metadata(metadata)
    return BrokerStartup(
        endpoint=_parse_broker_url(metadata["endpoint"]),
        process=process,
    )


def _mosquitto_available() -> bool:
    return bool(shutil.which("mosquitto"))


def _find_free_port(start_port: int) -> int:
    port = start_port
    while True:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            try:
                sock.bind(("127.0.0.1", port))
                return port
            except OSError:
                port += 1


def _publish_mqtt(
    endpoint: BrokerEndpoint, topic: str, envelope: GossipEnvelope
) -> None:
    try:
        import paho.mqtt.client as mqtt
    except ImportError:
        return
    client = mqtt.Client(client_id=producer_id())
    if endpoint.scheme == "mqtts":
        client.tls_set()
    client.connect(endpoint.host, endpoint.port, 30)
    client.loop_start()
    payload = envelope.model_dump(by_alias=True, mode="json")
    result = client.publish(topic, json.dumps(payload))
    result.wait_for_publish(timeout=2.0)
    client.loop_stop()
    client.disconnect()


def run_mqtt_subscription(
    endpoint: BrokerEndpoint,
    topics: list[str],
    handler: Callable[[GossipEnvelope], None],
) -> None:
    try:
        import paho.mqtt.client as mqtt
    except ImportError as exc:
        raise GossipError("paho-mqtt is required for MQTT transport") from exc

    client = mqtt.Client(client_id=producer_id())
    if endpoint.scheme == "mqtts":
        client.tls_set()

    def on_message(
        _client: mqtt.Client, _userdata: object, msg: mqtt.MQTTMessage
    ) -> None:
        try:
            payload = json.loads(msg.payload.decode())
            envelope = GossipEnvelope.model_validate(payload)
        except (json.JSONDecodeError, ValidationError):
            return
        handler(envelope)

    client.on_message = on_message
    client.connect(endpoint.host, endpoint.port, 30)
    for topic in topics:
        client.subscribe(topic)
    client.loop_forever()


def _resolve_project_label(
    root: Path, project_dir: Path, configuration: ProjectConfiguration
) -> Optional[str]:
    labeled = resolve_labeled_projects(root)
    for project in labeled:
        if project.project_dir == project_dir:
            return project.label
    return configuration.project_key


def _resolve_project_dir(root: Path, label: str) -> Optional[Path]:
    for project in resolve_labeled_projects(root):
        if project.label == label:
            return project.project_dir
    return None


def _uds_socket_path(realtime: Optional[RealtimeConfig] = None) -> Path:
    if realtime and realtime.uds_socket_path:
        return Path(realtime.uds_socket_path)
    runtime_dir = os.environ.get("XDG_RUNTIME_DIR")
    if runtime_dir:
        return Path(runtime_dir) / "kanbus" / "bus.sock"
    return Path.home() / ".kanbus" / "run" / "bus.sock"


def _parse_broker_url(url: str) -> BrokerEndpoint:
    if "://" not in url:
        raise GossipError(f"invalid broker url: {url}")
    scheme, rest = url.split("://", 1)
    host_port = rest.split("/", 1)[0]
    if ":" in host_port:
        host, port_str = host_port.split(":", 1)
        port = int(port_str)
    else:
        host = host_port
        port = 1883
    return BrokerEndpoint(scheme=scheme, host=host, port=port, url=url)


def _broker_run_dir() -> Path:
    return Path.home() / ".kanbus" / "run"


def _load_broker_metadata() -> dict | None:
    metadata_path = _broker_run_dir() / "broker.json"
    if not metadata_path.exists():
        return None
    try:
        return json.loads(metadata_path.read_text(encoding="utf-8"))
    except json.JSONDecodeError:
        return None


def _write_broker_metadata(metadata: dict) -> None:
    run_dir = _broker_run_dir()
    run_dir.mkdir(parents=True, exist_ok=True)
    path = run_dir / "broker.json"
    path.write_text(json.dumps(metadata, indent=2), encoding="utf-8")


def _now_iso() -> str:
    return (
        datetime.now(timezone.utc)
        .isoformat(timespec="milliseconds")
        .replace("+00:00", "Z")
    )


def _print_mosquitto_missing() -> None:
    print(
        "Mosquitto not found. Install with: brew install mosquitto (macOS) "
        "or apt install mosquitto (Debian/Ubuntu).",
        file=sys.stderr,
    )



from __future__ import annotations

from datetime import datetime, timezone
from pathlib import Path
from types import SimpleNamespace

import pytest

from kanbus import issue_comment
from kanbus.models import IssueComment

from test_helpers import build_issue


def _c(text: str, cid: str | None = None, author: str = "dev") -> IssueComment:
    return IssueComment.model_validate(
        {
            "id": cid,
            "author": author,
            "text": text,
            "created_at": datetime(2026, 3, 9, tzinfo=timezone.utc).isoformat(),
        }
    )


def test_normalize_prefix_and_find_comment_index() -> None:
    with pytest.raises(issue_comment.IssueCommentError, match="comment id is required"):
        issue_comment._normalize_prefix("  ")
    assert issue_comment._normalize_prefix(" AbC ") == "abc"

    issue = build_issue("kanbus-1")
    issue.comments = [_c("a", "abcdef"), _c("b", "abc999"), _c("c", "fff111")]

    with pytest.raises(issue_comment.IssueCommentError, match="ambiguous"):
        issue_comment._find_comment_index(issue, "abc")
    with pytest.raises(issue_comment.IssueCommentError, match="comment not found"):
        issue_comment._find_comment_index(issue, "zzz")
    assert issue_comment._find_comment_index(issue, "fff") == 2


def test_generate_comment_id_shape() -> None:
    generated = issue_comment._generate_comment_id()
    assert isinstance(generated, str)
    assert len(generated) >= 8


def test_ensure_comment_ids_changed_and_unchanged(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    issue = build_issue("kanbus-1")
    issue.comments = [_c("a", None), _c("b", "bb")]
    monkeypatch.setattr(issue_comment, "_generate_comment_id", lambda: "gen-1")

    updated, changed = issue_comment._ensure_comment_ids(issue)
    assert changed is True
    assert updated.comments[0].id == "gen-1"

    unchanged, changed2 = issue_comment._ensure_comment_ids(updated)
    assert changed2 is False
    assert unchanged is updated


def test_add_comment_success_and_failures(monkeypatch: pytest.MonkeyPatch) -> None:
    root = Path("/repo")
    project_dir = Path("/repo/project")
    issue_path = project_dir / "issues" / "kanbus-1.json"
    issue = build_issue("kanbus-1")
    issue.comments = [_c("old", None)]
    lookup = SimpleNamespace(
        issue=issue, issue_path=issue_path, project_dir=project_dir
    )

    monkeypatch.setattr(issue_comment, "load_issue_from_project", lambda *_a: lookup)
    ids = iter(["existing-id", "new-id"])
    monkeypatch.setattr(issue_comment, "_generate_comment_id", lambda: next(ids))
    writes: list[object] = []
    monkeypatch.setattr(
        issue_comment, "write_issue_to_file", lambda *_a: writes.append("w")
    )
    monkeypatch.setattr(
        issue_comment, "now_timestamp", lambda: "2026-03-09T00:00:00.000Z"
    )
    monkeypatch.setattr(issue_comment, "get_current_user", lambda: "dev")
    monkeypatch.setattr(
        issue_comment, "comment_payload", lambda _id, _author: {"ok": True}
    )
    monkeypatch.setattr(
        issue_comment,
        "create_event",
        lambda **_kwargs: SimpleNamespace(event_id="evt-1"),
    )
    monkeypatch.setattr(
        issue_comment, "events_dir_for_issue_path", lambda *_a: Path("/events")
    )
    monkeypatch.setattr(issue_comment, "write_events_batch", lambda *_a: None)
    published: list[str] = []
    monkeypatch.setattr(
        issue_comment, "publish_issue_mutation", lambda *_a: published.append("pub")
    )

    result = issue_comment.add_comment(root, "kanbus-1", "dev", "new")
    assert result.comment.id == "new-id"
    assert len(result.issue.comments) == 2
    assert published == ["pub"]

    monkeypatch.setattr(
        issue_comment,
        "load_issue_from_project",
        lambda *_a: (_ for _ in ()).throw(issue_comment.IssueLookupError("missing")),
    )
    with pytest.raises(issue_comment.IssueCommentError, match="missing"):
        issue_comment.add_comment(root, "x", "dev", "x")

    monkeypatch.setattr(issue_comment, "load_issue_from_project", lambda *_a: lookup)
    monkeypatch.setattr(issue_comment, "_generate_comment_id", lambda: "id-x")
    monkeypatch.setattr(
        issue_comment,
        "write_events_batch",
        lambda *_a: (_ for _ in ()).throw(RuntimeError("event fail")),
    )
    with pytest.raises(issue_comment.IssueCommentError, match="event fail"):
        issue_comment.add_comment(root, "kanbus-1", "dev", "new")
    assert len(writes) >= 2

    monkeypatch.setattr(issue_comment, "_generate_comment_id", lambda: None)
    monkeypatch.setattr(issue_comment, "write_events_batch", lambda *_a: None)
    with pytest.raises(issue_comment.IssueCommentError, match="comment id is required"):
        issue_comment.add_comment(root, "kanbus-1", "dev", "new")


def test_ensure_issue_comment_ids(monkeypatch: pytest.MonkeyPatch) -> None:
    issue = build_issue("kanbus-1")
    issue.comments = [_c("old", None)]
    lookup = SimpleNamespace(
        issue=issue, issue_path=Path("/repo/project/issues/kanbus-1.json")
    )
    monkeypatch.setattr(issue_comment, "load_issue_from_project", lambda *_a: lookup)
    monkeypatch.setattr(issue_comment, "_generate_comment_id", lambda: "id-1")
    wrote: list[str] = []
    monkeypatch.setattr(
        issue_comment, "write_issue_to_file", lambda *_a: wrote.append("w")
    )

    updated = issue_comment.ensure_issue_comment_ids(Path("/repo"), "kanbus-1")
    assert updated.comments[0].id == "id-1"
    assert wrote == ["w"]

    monkeypatch.setattr(
        issue_comment,
        "load_issue_from_project",
        lambda *_a: (_ for _ in ()).throw(issue_comment.IssueLookupError("missing")),
    )
    with pytest.raises(issue_comment.IssueCommentError, match="missing"):
        issue_comment.ensure_issue_comment_ids(Path("/repo"), "x")


def test_update_comment_success_and_paths(monkeypatch: pytest.MonkeyPatch) -> None:
    root = Path("/repo")
    project_dir = Path("/repo/project")
    issue_path = project_dir / "issues" / "kanbus-1.json"
    issue = build_issue("kanbus-1")
    issue.comments = [_c("old", "abc123")]
    lookup = SimpleNamespace(
        issue=issue, issue_path=issue_path, project_dir=project_dir
    )

    monkeypatch.setattr(issue_comment, "load_issue_from_project", lambda *_a: lookup)
    monkeypatch.setattr(issue_comment, "write_issue_to_file", lambda *_a: None)
    monkeypatch.setattr(
        issue_comment, "now_timestamp", lambda: "2026-03-09T00:00:00.000Z"
    )
    monkeypatch.setattr(issue_comment, "get_current_user", lambda: "dev")
    monkeypatch.setattr(
        issue_comment, "comment_updated_payload", lambda *_a: {"ok": True}
    )
    monkeypatch.setattr(
        issue_comment,
        "create_event",
        lambda **_kwargs: SimpleNamespace(event_id="evt-2"),
    )
    monkeypatch.setattr(
        issue_comment, "events_dir_for_issue_path", lambda *_a: Path("/events")
    )
    monkeypatch.setattr(issue_comment, "write_events_batch", lambda *_a: None)
    monkeypatch.setattr(issue_comment, "publish_issue_mutation", lambda *_a: None)

    updated = issue_comment.update_comment(root, "kanbus-1", "abc", "new text")
    assert updated.comments[0].text == "new text"

    monkeypatch.setattr(
        issue_comment,
        "load_issue_from_project",
        lambda *_a: (_ for _ in ()).throw(issue_comment.IssueLookupError("missing")),
    )
    with pytest.raises(issue_comment.IssueCommentError, match="missing"):
        issue_comment.update_comment(root, "x", "abc", "new")

    monkeypatch.setattr(issue_comment, "load_issue_from_project", lambda *_a: lookup)
    no_id_issue = build_issue("kanbus-1")
    no_id_issue.comments = [_c("old", None)]
    monkeypatch.setattr(
        issue_comment, "_ensure_comment_ids", lambda _issue: (no_id_issue, False)
    )
    monkeypatch.setattr(issue_comment, "_find_comment_index", lambda *_a: 0)
    with pytest.raises(issue_comment.IssueCommentError, match="comment id is required"):
        issue_comment.update_comment(root, "kanbus-1", "abc", "new")


def test_update_delete_comment_event_failure_rolls_back(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    root = Path("/repo")
    project_dir = Path("/repo/project")
    issue_path = project_dir / "local" / "kanbus-1.json"
    issue = build_issue("kanbus-1")
    issue.comments = [_c("old", "abc123")]
    lookup = SimpleNamespace(
        issue=issue, issue_path=issue_path, project_dir=project_dir
    )

    monkeypatch.setattr(issue_comment, "load_issue_from_project", lambda *_a: lookup)
    writes: list[str] = []
    monkeypatch.setattr(
        issue_comment, "write_issue_to_file", lambda *_a: writes.append("w")
    )
    monkeypatch.setattr(
        issue_comment, "now_timestamp", lambda: "2026-03-09T00:00:00.000Z"
    )
    monkeypatch.setattr(issue_comment, "get_current_user", lambda: "dev")
    monkeypatch.setattr(
        issue_comment, "comment_updated_payload", lambda *_a: {"ok": True}
    )
    monkeypatch.setattr(issue_comment, "comment_payload", lambda *_a: {"ok": True})
    monkeypatch.setattr(
        issue_comment,
        "create_event",
        lambda **_kwargs: SimpleNamespace(event_id="evt-3"),
    )
    monkeypatch.setattr(
        issue_comment, "events_dir_for_issue_path", lambda *_a: Path("/events")
    )
    monkeypatch.setattr(
        issue_comment,
        "write_events_batch",
        lambda *_a: (_ for _ in ()).throw(RuntimeError("event fail")),
    )
    monkeypatch.setattr(
        issue_comment,
        "publish_issue_mutation",
        lambda *_a: (_ for _ in ()).throw(RuntimeError("should not publish")),
    )

    with pytest.raises(issue_comment.IssueCommentError, match="event fail"):
        issue_comment.update_comment(root, "kanbus-1", "abc", "new")
    with pytest.raises(issue_comment.IssueCommentError, match="event fail"):
        issue_comment.delete_comment(root, "kanbus-1", "abc")
    assert len(writes) >= 4


def test_delete_comment_success_lookup_error_and_missing_id(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    root = Path("/repo")
    project_dir = Path("/repo/project")
    issue_path = project_dir / "issues" / "kanbus-1.json"
    issue = build_issue("kanbus-1")
    issue.comments = [_c("old", "abc123")]
    lookup = SimpleNamespace(
        issue=issue, issue_path=issue_path, project_dir=project_dir
    )

    monkeypatch.setattr(issue_comment, "load_issue_from_project", lambda *_a: lookup)
    monkeypatch.setattr(issue_comment, "write_issue_to_file", lambda *_a: None)
    monkeypatch.setattr(
        issue_comment, "now_timestamp", lambda: "2026-03-09T00:00:00.000Z"
    )
    monkeypatch.setattr(issue_comment, "get_current_user", lambda: "dev")
    monkeypatch.setattr(issue_comment, "comment_payload", lambda *_a: {"ok": True})
    monkeypatch.setattr(
        issue_comment,
        "create_event",
        lambda **_kwargs: SimpleNamespace(event_id="evt-4"),
    )
    monkeypatch.setattr(
        issue_comment, "events_dir_for_issue_path", lambda *_a: Path("/events")
    )
    monkeypatch.setattr(issue_comment, "write_events_batch", lambda *_a: None)
    published: list[str] = []
    monkeypatch.setattr(
        issue_comment, "publish_issue_mutation", lambda *_a: published.append("pub")
    )

    updated = issue_comment.delete_comment(root, "kanbus-1", "abc")
    assert updated.comments == []
    assert published == ["pub"]

    monkeypatch.setattr(
        issue_comment,
        "load_issue_from_project",
        lambda *_a: (_ for _ in ()).throw(issue_comment.IssueLookupError("missing")),
    )
    with pytest.raises(issue_comment.IssueCommentError, match="missing"):
        issue_comment.delete_comment(root, "x", "abc")

    monkeypatch.setattr(issue_comment, "load_issue_from_project", lambda *_a: lookup)
    no_id_issue = build_issue("kanbus-1")
    no_id_issue.comments = [_c("old", None)]
    monkeypatch.setattr(
        issue_comment, "_ensure_comment_ids", lambda _issue: (no_id_issue, False)
    )
    monkeypatch.setattr(issue_comment, "_find_comment_index", lambda *_a: 0)
    with pytest.raises(issue_comment.IssueCommentError, match="comment id is required"):
        issue_comment.delete_comment(root, "kanbus-1", "abc")



from __future__ import annotations

import json
from pathlib import Path

from kanbus import cache, index
from kanbus.models import DependencyLink

from test_helpers import build_issue


def _write_issue(path: Path, issue_id: str, **kwargs: object) -> None:
    issue = build_issue(issue_id, **kwargs)
    path.write_text(
        json.dumps(issue.model_dump(by_alias=True, mode="json")), encoding="utf-8"
    )


def test_load_issue_data_and_batch(tmp_path: Path) -> None:
    issue_path = tmp_path / "kanbus-1.json"
    _write_issue(issue_path, "kanbus-1", issue_type="bug")

    loaded = index._load_issue_data(issue_path)
    batched = index._load_issue_batch([issue_path])

    assert loaded.identifier == "kanbus-1"
    assert batched[0].issue_type == "bug"


def test_add_issue_to_index_indexes_all_dimensions() -> None:
    issue = build_issue(
        "kanbus-1",
        issue_type="task",
        status="open",
        parent="kanbus-parent",
        labels=["backend"],
    )
    issue = issue.model_copy(
        update={
            "dependencies": [
                DependencyLink.model_validate(
                    {"target": "kanbus-base", "type": "blocked-by"}
                )
            ]
        }
    )
    issue_index = index.IssueIndex()

    index._add_issue_to_index(issue_index, issue)

    assert issue_index.by_id["kanbus-1"].identifier == "kanbus-1"
    assert issue_index.by_status["open"][0].identifier == "kanbus-1"
    assert issue_index.by_type["task"][0].identifier == "kanbus-1"
    assert issue_index.by_parent["kanbus-parent"][0].identifier == "kanbus-1"
    assert issue_index.by_label["backend"][0].identifier == "kanbus-1"
    assert issue_index.reverse_dependencies["kanbus-base"][0].identifier == "kanbus-1"


def test_build_index_from_directory_filters_and_loads(tmp_path: Path) -> None:
    _write_issue(
        tmp_path / "kanbus-2.json", "kanbus-2", issue_type="bug", status="closed"
    )
    _write_issue(
        tmp_path / "kanbus-1.json", "kanbus-1", issue_type="task", status="open"
    )
    (tmp_path / "ignore.txt").write_text("x", encoding="utf-8")

    issue_index = index.build_index_from_directory(tmp_path)

    assert set(issue_index.by_id.keys()) == {"kanbus-1", "kanbus-2"}
    assert [i.identifier for i in issue_index.by_status["open"]] == ["kanbus-1"]
    assert [i.identifier for i in issue_index.by_type["bug"]] == ["kanbus-2"]


def test_collect_issue_file_mtimes_ignores_non_json(tmp_path: Path) -> None:
    json_path = tmp_path / "kanbus-1.json"
    json_path.write_text("{}", encoding="utf-8")
    (tmp_path / "ignore.md").write_text("x", encoding="utf-8")

    mtimes = cache.collect_issue_file_mtimes(tmp_path)
    assert set(mtimes.keys()) == {"kanbus-1.json"}


def test_normalize_mtime_rounds_to_microseconds() -> None:
    assert cache._normalize_mtime(1.12345678) == 1.123457


def test_load_cache_if_valid_missing_or_stale(tmp_path: Path) -> None:
    cache_path = tmp_path / "index.json"

    assert cache.load_cache_if_valid(cache_path, tmp_path) is None

    cache_path.write_text(
        json.dumps({"file_mtimes": {"kanbus-1.json": 1.0}, "issues": []}),
        encoding="utf-8",
    )
    assert cache.load_cache_if_valid(cache_path, tmp_path) is None


def test_write_cache_and_reload(tmp_path: Path) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()

    issue = build_issue("kanbus-1", status="open", issue_type="task", labels=["l1"])
    issue_index = index.IssueIndex(
        by_id={issue.identifier: issue},
        by_status={"open": [issue]},
        by_type={"task": [issue]},
        by_label={"l1": [issue]},
        reverse_dependencies={"kanbus-x": [issue]},
    )
    mtimes = {"kanbus-1.json": 123.456789}
    cache_path = tmp_path / ".cache" / "index.json"

    cache.write_cache(issue_index, cache_path, mtimes)
    loaded = cache.load_cache_if_valid(cache_path, issues_dir)

    assert loaded is None

    issue_file = issues_dir / "kanbus-1.json"
    issue_file.write_text(
        json.dumps(issue.model_dump(by_alias=True, mode="json")), encoding="utf-8"
    )
    current_mtimes = cache.collect_issue_file_mtimes(issues_dir)
    cache.write_cache(issue_index, cache_path, current_mtimes)

    loaded = cache.load_cache_if_valid(cache_path, issues_dir)
    assert loaded is not None
    assert loaded.by_id["kanbus-1"].identifier == "kanbus-1"
    assert loaded.reverse_dependencies["kanbus-x"][0].identifier == "kanbus-1"


def test_build_index_from_cache_ignores_unknown_reverse_dependency_ids() -> None:
    issue = build_issue("kanbus-1", issue_type="task", status="open", parent="kanbus-0")

    rebuilt = cache.build_index_from_cache(
        [issue], {"kanbus-x": ["kanbus-1", "missing-id"]}
    )

    assert rebuilt.by_id["kanbus-1"].identifier == "kanbus-1"
    assert rebuilt.by_parent["kanbus-0"][0].identifier == "kanbus-1"
    assert [item.identifier for item in rebuilt.reverse_dependencies["kanbus-x"]] == [
        "kanbus-1"
    ]



"""Wiki rendering utilities."""

from __future__ import annotations

import hashlib
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List

from jinja2 import Environment, FileSystemLoader, select_autoescape

from kanbus.ai_summarize import make_ai_summarize
from kanbus.console_snapshot import ConsoleSnapshotError, get_issues_for_root
from kanbus.models import IssueData
from kanbus.queries import filter_issues


class WikiError(RuntimeError):
    """Raised when wiki rendering fails."""


@dataclass(frozen=True)
class WikiContext:
    """Wiki context for rendering.

    :param issues: Issues loaded for rendering.
    :type issues: List[IssueData]
    """

    issues: List[IssueData]

    def query(self, **filters: object) -> List[Dict[str, object]]:
        """Query issues for wiki templates.

        :return: Matching issues.
        :rtype: List[Dict[str, object]]
        :raises WikiError: If the query parameters are invalid.
        """
        status = _get_string(filters.get("status"))
        issue_type = _get_string(filters.get("issue_type") or filters.get("type"))
        sort_key = _get_string(filters.get("sort"))

        filtered = filter_issues(
            self.issues,
            status,
            issue_type,
            None,
            None,
        )
        if sort_key is None:
            return [_serialize_issue(issue) for issue in filtered]
        if sort_key == "title":
            return [
                _serialize_issue(issue)
                for issue in sorted(filtered, key=lambda issue: issue.title)
            ]
        if sort_key == "priority":
            return [
                _serialize_issue(issue)
                for issue in sorted(filtered, key=lambda issue: issue.priority)
            ]
        raise WikiError("invalid sort key")

    def count(self, **filters: object) -> int:
        """Count issues for wiki templates.

        :return: Count of matching issues.
        :rtype: int
        :raises WikiError: If the query parameters are invalid.
        """
        return len(self.query(**filters))

    def issue(self, identifier: str) -> Dict[str, object] | None:
        """Look up an issue by identifier for wiki templates.

        :param identifier: Issue identifier.
        :type identifier: str
        :return: Serialized issue or None if not found.
        :rtype: Dict[str, object] | None
        """
        for i in self.issues:
            if i.identifier == identifier:
                return _serialize_issue(i)
        return None


@dataclass(frozen=True)
class WikiRenderRequest:
    """Request for rendering a wiki page.

    :param root: Repository root path.
    :type root: Path
    :param page_path: Path to the wiki page.
    :type page_path: Path
    """

    root: Path
    page_path: Path


def render_wiki_page(request: WikiRenderRequest) -> str:
    """Render a wiki page using the live issue index.

    :param request: Render request with root and page path.
    :type request: WikiRenderRequest
    :return: Rendered wiki content.
    :rtype: str
    :raises WikiError: If rendering fails.
    """
    full_page = request.root / request.page_path
    if not full_page.exists():
        raise WikiError("wiki page not found")

    try:
        issues = get_issues_for_root(request.root)
    except ConsoleSnapshotError as error:
        raise WikiError(str(error)) from error

    ai_config, project_dir = _load_ai_config_and_project_dir(request.root)
    wiki_render_cache_dir = (
        request.root / project_dir / ".cache" / "wiki_render" if project_dir else None
    )
    if wiki_render_cache_dir is not None:
        cache_key = _wiki_render_cache_key(full_page, list(issues))
        cached = _wiki_render_read_cache(wiki_render_cache_dir, cache_key)
        if cached is not None:
            _wiki_render_log_cache_hit(wiki_render_cache_dir)
            return cached
    issues_by_id = {i.identifier: _serialize_issue(i) for i in issues}
    ai_cache_dir = request.root / project_dir / ".cache" if project_dir else None
    ai_summarize_fn = make_ai_summarize(issues_by_id, ai_config, ai_cache_dir)

    context = WikiContext(issues=list(issues))
    environment = Environment(
        loader=FileSystemLoader(str(request.page_path.parent)),
        autoescape=select_autoescape(
            enabled_extensions=("html", "htm", "xml"),
            default_for_string=False,
            default=False,
        ),
    )
    environment.globals.update(
        {
            "query": context.query,
            "count": context.count,
            "issue": context.issue,
            "ai_summarize": ai_summarize_fn,
        }
    )
    try:
        rendered = environment.get_template(request.page_path.name).render()
    except WikiError:
        raise
    except Exception as error:
        raise WikiError(str(error)) from error

    if wiki_render_cache_dir is not None:
        _wiki_render_write_cache(wiki_render_cache_dir, cache_key, rendered)
    return rendered


def _wiki_render_cache_key(page_path: Path, issues: List[IssueData]) -> str:
    page_mtime = str(page_path.stat().st_mtime) if page_path.exists() else ""
    issue_part = "|".join(
        f"{i.identifier}:{i.updated_at.isoformat()}"
        for i in sorted(issues, key=lambda x: x.identifier)
    )
    raw = f"{page_path}|{page_mtime}|{issue_part}"
    return hashlib.sha256(raw.encode()).hexdigest()


def _wiki_render_read_cache(cache_dir: Path, key: str) -> str | None:
    path = cache_dir / f"{key}.md"
    if not path.exists():
        return None
    try:
        return path.read_text(encoding="utf-8")
    except OSError:
        return None


def _wiki_render_write_cache(cache_dir: Path, key: str, content: str) -> None:
    cache_dir.mkdir(parents=True, exist_ok=True)
    (cache_dir / f"{key}.md").write_text(content, encoding="utf-8")


def _wiki_render_log_cache_hit(cache_dir: Path) -> None:
    log_path = cache_dir.parent / "wiki_cache_hits.log"
    cache_dir.parent.mkdir(parents=True, exist_ok=True)
    log_path.open("a", encoding="utf-8").write("1\n")


def _load_ai_config_and_project_dir(root: Path) -> tuple[object | None, str | None]:
    """Load AI configuration and project directory. Returns (ai_config, project_dir)."""
    from kanbus.config_loader import ConfigurationError, load_project_configuration
    from kanbus.project import ProjectMarkerError, get_configuration_path

    try:
        config_path = get_configuration_path(root)
        configuration = load_project_configuration(config_path)
        return (configuration.ai, configuration.project_directory)
    except (ProjectMarkerError, ConfigurationError):
        return (None, None)


def _get_string(value: object) -> str | None:
    if value is None:
        return None
    if isinstance(value, str):
        return value
    raise WikiError("invalid query parameter")


def _serialize_issue(issue: IssueData) -> Dict[str, object]:
    return issue.model_dump(by_alias=True, mode="json")


def render_template_string(text: str, issues: List[IssueData]) -> str:
    """Render a template string with wiki context (query, count, issue).

    :param text: Template string (may contain Jinja2).
    :type text: str
    :param issues: Issues for query/count/issue context.
    :type issues: List[IssueData]
    :return: Rendered text.
    :rtype: str
    :raises WikiError: If template rendering fails.
    """
    context = WikiContext(issues=issues)
    environment = Environment(
        autoescape=select_autoescape(
            enabled_extensions=("html", "htm", "xml"),
            default_for_string=False,
            default=False,
        ),
    )
    environment.globals.update(
        {
            "query": context.query,
            "count": context.count,
            "issue": context.issue,
        }
    )
    try:
        template = environment.from_string(text)
        return template.render()
    except WikiError:
        raise
    except Exception as error:
        raise WikiError(str(error)) from error


def list_wiki_pages(root: Path) -> List[str]:
    """List wiki page paths relative to repository root.

    :param root: Repository root path.
    :type root: Path
    :return: Sorted list of paths like project/docs/page.md.
    :rtype: List[str]
    :raises WikiError: If configuration or project structure is invalid.
    """
    from kanbus.config_loader import ConfigurationError, load_project_configuration
    from kanbus.project import ProjectMarkerError, get_configuration_path

    try:
        config_path = get_configuration_path(root)
        configuration = load_project_configuration(config_path)
    except (ProjectMarkerError, ConfigurationError) as error:
        raise WikiError(str(error)) from error

    project_dir = configuration.project_directory
    wiki_subdir = configuration.wiki_directory or "wiki"
    if wiki_subdir.startswith("../"):
        normalized = wiki_subdir.replace("\\", "/").lstrip("../").lstrip("..\\")
        wiki_root = root / normalized
        prefix = Path(normalized)
    else:
        wiki_root = root / project_dir / wiki_subdir
        prefix = Path(project_dir) / wiki_subdir

    if not wiki_root.exists():
        return []

    paths: List[str] = []
    for path in wiki_root.rglob("*.md"):
        if path.is_file():
            rel = path.relative_to(wiki_root)
            paths.append(str(prefix / rel).replace("\\", "/"))
    paths.sort()
    return paths



"""User identification helpers."""

from __future__ import annotations

import os


def get_current_user() -> str:
    """Return the current user identifier.

    :return: Current user identifier.
    :rtype: str
    """
    override = os.getenv("KANBUS_USER")
    if override is not None and override.strip():
        return override
    user = os.getenv("USER")
    if user:
        return user
    return "unknown"



"""Daemon entry point."""

from __future__ import annotations

import argparse
from pathlib import Path

from kanbus.daemon_server import run_daemon


def parse_args(argv: list[str]) -> argparse.Namespace:
    """Parse daemon CLI arguments.

    :param argv: Command line arguments.
    :type argv: list[str]
    :return: Parsed arguments.
    :rtype: argparse.Namespace
    """
    parser = argparse.ArgumentParser(description="Kanbus daemon")
    parser.add_argument("--root", required=True)
    return parser.parse_args(argv)


def main(argv: list[str]) -> None:
    """Run the daemon entry point.

    :param argv: Command line arguments.
    :type argv: list[str]
    """
    args = parse_args(argv)
    run_daemon(Path(args.root))


if __name__ == "__main__":
    import sys

    main(sys.argv[1:])



"""GitHub security synchronization support.

Pulls Dependabot alerts from the GitHub REST API and creates/updates Kanbus
issues under a managed hierarchy.
"""

from __future__ import annotations

import os
import logging
import subprocess
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Set

import requests

from kanbus.ids import IssueIdentifierRequest, generate_issue_identifier
from kanbus.issue_files import (
    list_issue_identifiers,
    read_issue_from_file,
    write_issue_to_file,
)
from kanbus.beads_write import create_beads_issue, update_beads_issue
from kanbus.migration import load_beads_issues
from kanbus.models import (
    DependabotConfiguration,
    GithubSecurityConfiguration,
    IssueData,
)

GITHUB_API_BASE = "https://api.github.com"
GITHUB_API_VERSION = "2022-11-28"
GITHUB_SECURITY_INITIATIVE_TITLE = "GitHub Security Remediation"
GITHUB_DEPENDABOT_EPIC_TITLE = "GitHub Dependabot Alerts"
LOGGER = logging.getLogger(__name__)


class GithubSecuritySyncError(RuntimeError):
    """Raised when a GitHub security sync operation fails."""


@dataclass
class DependabotPullResult:
    """Result of a Dependabot pull operation."""

    pulled: int = 0
    updated: int = 0
    skipped: int = 0


def pull_dependabot_from_github(
    root: Path,
    github_security_config: GithubSecurityConfiguration,
    project_key: str,
    dry_run: bool = False,
) -> DependabotPullResult:
    """Pull Dependabot alerts from GitHub and create/update Kanbus issues."""
    token = os.environ.get("GITHUB_TOKEN") or os.environ.get("GH_TOKEN")
    if not token:
        raise GithubSecuritySyncError(
            "GITHUB_TOKEN or GH_TOKEN environment variable is not set"
        )

    dependabot_config = github_security_config.dependabot or DependabotConfiguration()
    _validate_state(dependabot_config.state)
    min_priority = _severity_to_priority(dependabot_config.min_severity)

    repo = github_security_config.repo or _detect_repo_from_git(root)
    if not repo:
        raise GithubSecuritySyncError(
            "could not determine GitHub repository slug "
            "(use --repo or github_security.repo)"
        )

    from kanbus.project import load_project_directory

    project_dir = load_project_directory(root)
    issues_dir = project_dir / "issues"
    if not issues_dir.exists():
        raise GithubSecuritySyncError("issues directory does not exist")

    alerts = _fetch_dependabot_alerts(repo, token, dependabot_config.state)
    alerts = [
        alert
        for alert in alerts
        if _severity_to_priority(_alert_severity(alert)) <= min_priority
    ]

    all_existing: Set[str] = set(list_issue_identifiers(issues_dir))
    alert_index = _build_alert_index(all_existing, issues_dir)
    task_index = _build_task_index(all_existing, issues_dir)

    parent_epic = _resolve_dependabot_epic(
        issues_dir,
        project_key,
        dependabot_config.parent_epic,
        dry_run,
        all_existing,
    )

    grouped: Dict[str, List[Dict[str, Any]]] = defaultdict(list)
    for alert in alerts:
        grouped[_task_target_key(alert)].append(alert)

    result = DependabotPullResult()

    for target_key, alerts_for_target in sorted(grouped.items()):
        task_id = _resolve_manifest_task(
            issues_dir,
            project_key,
            repo,
            target_key,
            parent_epic,
            _min_priority(alerts_for_target),
            dry_run,
            task_index,
            all_existing,
        )

        for alert in alerts_for_target:
            alert_number = _alert_number(alert)
            if alert_number <= 0:
                continue

            index_key = f"{repo}#{alert_number}"
            existing_kanbus_id = alert_index.get(index_key)
            if existing_kanbus_id:
                kanbus_id = existing_kanbus_id
                action = "updated"
            else:
                request = IssueIdentifierRequest(
                    title=_dependabot_alert_title(alert),
                    existing_ids=frozenset(all_existing),
                    prefix=project_key,
                )
                identifier = generate_issue_identifier(request).identifier
                all_existing.add(identifier)
                kanbus_id = identifier
                action = "pulled "

            issue = _map_dependabot_to_kanbus(alert, repo, task_id)
            issue = issue.model_copy(update={"identifier": kanbus_id})

            issue_path = _issue_path(issues_dir, kanbus_id)
            if action == "updated" and issue_path.exists():
                try:
                    existing = read_issue_from_file(issue_path)
                    issue = issue.model_copy(update={"created_at": existing.created_at})
                except Exception:
                    # Keep pull flow best-effort; continue while surfacing failure details.
                    LOGGER.exception(
                        "Failed to preserve created_at from existing issue at %s",
                        issue_path,
                    )

            severity = _alert_severity(alert)
            print(
                f'{action}  [{severity:<8}]  {"alert#" + str(alert_number):<14}  "{issue.title}"'
            )

            if not dry_run:
                write_issue_to_file(issue, issue_path)

            if action == "updated":
                result.updated += 1
            else:
                result.pulled += 1

    return result


def pull_dependabot_from_github_beads(
    root: Path,
    github_security_config: GithubSecurityConfiguration,
    dry_run: bool = False,
) -> DependabotPullResult:
    """Pull Dependabot alerts from GitHub and create/update Beads issues."""
    token = os.environ.get("GITHUB_TOKEN") or os.environ.get("GH_TOKEN")
    if not token:
        raise GithubSecuritySyncError(
            "GITHUB_TOKEN or GH_TOKEN environment variable is not set"
        )

    dependabot_config = github_security_config.dependabot or DependabotConfiguration()
    _validate_state(dependabot_config.state)
    min_priority = _severity_to_priority(dependabot_config.min_severity)

    repo = github_security_config.repo or _detect_repo_from_git(root)
    if not repo:
        raise GithubSecuritySyncError(
            "could not determine GitHub repository slug "
            "(use --repo or github_security.repo)"
        )

    alerts = _fetch_dependabot_alerts(repo, token, dependabot_config.state)
    alerts = [
        alert
        for alert in alerts
        if _severity_to_priority(_alert_severity(alert)) <= min_priority
    ]

    existing_issues = load_beads_issues(root)
    alert_index = _build_beads_alert_index(existing_issues)
    task_index = _build_beads_task_index(existing_issues)

    initiative_id = _resolve_beads_initiative(root, existing_issues, dry_run)
    parent_epic = _resolve_beads_epic(
        root,
        existing_issues,
        dependabot_config.parent_epic,
        initiative_id,
        dry_run,
    )

    grouped: Dict[str, List[Dict[str, Any]]] = defaultdict(list)
    for alert in alerts:
        grouped[_task_target_key(alert)].append(alert)

    result = DependabotPullResult()
    runtime_alert_index = dict(alert_index)
    runtime_task_index = dict(task_index)

    for target_key, alerts_for_target in sorted(grouped.items()):
        task_id = _resolve_beads_task(
            root=root,
            repository=repo,
            target_key=target_key,
            parent_epic=parent_epic,
            priority=_min_priority(alerts_for_target),
            dry_run=dry_run,
            task_index=runtime_task_index,
        )

        for alert in alerts_for_target:
            alert_number = _alert_number(alert)
            if alert_number <= 0:
                continue

            index_key = f"{repo}#{alert_number}"
            existing_id = runtime_alert_index.get(index_key)
            title = _dependabot_alert_title(alert)
            description = _map_dependabot_to_beads_description(alert, repo)
            severity = _alert_severity(alert)

            if existing_id:
                if not dry_run:
                    update_beads_issue(
                        root=root,
                        identifier=existing_id,
                        status="open",
                        title=title,
                        description=description,
                        priority=_severity_to_priority(severity),
                        set_labels=["security", "github", "dependabot"],
                    )
                action = "updated"
            else:
                action = "pulled "
                if dry_run:
                    pass
                else:
                    created = create_beads_issue(
                        root=root,
                        title=title,
                        issue_type="sub-task",
                        priority=_severity_to_priority(severity),
                        assignee=None,
                        parent=task_id,
                        description=description,
                    )
                    update_beads_issue(
                        root=root,
                        identifier=created.identifier,
                        status="open",
                        title=title,
                        description=description,
                        priority=_severity_to_priority(severity),
                        set_labels=["security", "github", "dependabot"],
                    )
                    runtime_alert_index[index_key] = created.identifier

            print(
                f'{action}  [{severity:<8}]  {"alert#" + str(alert_number):<14}  "{title}"'
            )
            if action == "updated":
                result.updated += 1
            else:
                result.pulled += 1

    return result


def _validate_state(state: str) -> None:
    allowed = {"open", "fixed", "dismissed", "auto_dismissed"}
    if state not in allowed:
        raise GithubSecuritySyncError(
            f"invalid dependabot state '{state}' "
            f"(expected one of: {', '.join(sorted(allowed))})"
        )


def _resolve_dependabot_epic(
    issues_dir: Path,
    project_key: str,
    configured_id: Optional[str],
    dry_run: bool,
    all_existing: Set[str],
) -> str:
    if configured_id and _issue_path(issues_dir, configured_id).exists():
        return configured_id

    initiative_id = _resolve_security_initiative(
        issues_dir, project_key, dry_run, all_existing
    )
    existing = _find_existing_dependabot_epic(issues_dir, all_existing, initiative_id)
    if existing:
        return existing

    request = IssueIdentifierRequest(
        title=GITHUB_DEPENDABOT_EPIC_TITLE,
        existing_ids=frozenset(all_existing),
        prefix=project_key,
    )
    epic_id = generate_issue_identifier(request).identifier
    all_existing.add(epic_id)

    now = datetime.now(timezone.utc)
    epic = IssueData.model_validate(
        {
            "id": epic_id,
            "title": GITHUB_DEPENDABOT_EPIC_TITLE,
            "description": "Dependabot alerts imported from GitHub Security.",
            "type": "epic",
            "status": "open",
            "priority": 1,
            "assignee": None,
            "creator": None,
            "parent": initiative_id,
            "labels": ["security", "github", "dependabot"],
            "dependencies": [],
            "comments": [],
            "created_at": now.isoformat(),
            "updated_at": now.isoformat(),
            "closed_at": None,
            "custom": {},
        }
    )
    print(f'created  [epic    ]  "{GITHUB_DEPENDABOT_EPIC_TITLE}"')
    if not dry_run:
        write_issue_to_file(epic, _issue_path(issues_dir, epic_id))
    return epic_id


def _resolve_security_initiative(
    issues_dir: Path,
    project_key: str,
    dry_run: bool,
    all_existing: Set[str],
) -> str:
    existing = _find_existing_security_initiative(issues_dir, all_existing)
    if existing:
        return existing

    request = IssueIdentifierRequest(
        title=GITHUB_SECURITY_INITIATIVE_TITLE,
        existing_ids=frozenset(all_existing),
        prefix=project_key,
    )
    initiative_id = generate_issue_identifier(request).identifier
    all_existing.add(initiative_id)

    now = datetime.now(timezone.utc)
    initiative = IssueData.model_validate(
        {
            "id": initiative_id,
            "title": GITHUB_SECURITY_INITIATIVE_TITLE,
            "description": "Track remediation of GitHub security findings.",
            "type": "initiative",
            "status": "open",
            "priority": 1,
            "assignee": None,
            "creator": None,
            "parent": None,
            "labels": ["security", "github"],
            "dependencies": [],
            "comments": [],
            "created_at": now.isoformat(),
            "updated_at": now.isoformat(),
            "closed_at": None,
            "custom": {},
        }
    )
    print(f'created  [initiative]  "{GITHUB_SECURITY_INITIATIVE_TITLE}"')
    if not dry_run:
        write_issue_to_file(initiative, _issue_path(issues_dir, initiative_id))
    return initiative_id


def _find_existing_security_initiative(
    issues_dir: Path, all_existing: Set[str]
) -> Optional[str]:
    best: Optional[IssueData] = None
    for identifier in all_existing:
        try:
            issue = read_issue_from_file(_issue_path(issues_dir, identifier))
        except Exception:
            continue
        if issue.issue_type != "initiative":
            continue
        if issue.title != GITHUB_SECURITY_INITIATIVE_TITLE:
            continue
        if "github" not in issue.labels:
            continue
        if best is None or issue.updated_at > best.updated_at:
            best = issue
    return best.identifier if best else None


def _find_existing_dependabot_epic(
    issues_dir: Path, all_existing: Set[str], parent_initiative: str
) -> Optional[str]:
    best: Optional[IssueData] = None
    for identifier in all_existing:
        try:
            issue = read_issue_from_file(_issue_path(issues_dir, identifier))
        except Exception:
            continue
        if issue.issue_type != "epic":
            continue
        if issue.title != GITHUB_DEPENDABOT_EPIC_TITLE:
            continue
        if "dependabot" not in issue.labels:
            continue
        if issue.parent != parent_initiative:
            continue
        if best is None or issue.updated_at > best.updated_at:
            best = issue
    return best.identifier if best else None


def _resolve_manifest_task(
    issues_dir: Path,
    project_key: str,
    repo: str,
    target_key: str,
    parent_epic: str,
    priority: int,
    dry_run: bool,
    task_index: Dict[str, str],
    all_existing: Set[str],
) -> str:
    existing = task_index.get(target_key)
    if existing:
        if not dry_run:
            path = _issue_path(issues_dir, existing)
            task = read_issue_from_file(path)
            changed = False
            if task.parent != parent_epic:
                task.parent = parent_epic
                changed = True
            if task.priority != priority:
                task.priority = priority
                changed = True
            for label in ("security", "github", "dependabot"):
                if label not in task.labels:
                    task.labels.append(label)
                    changed = True
            if changed:
                task.updated_at = datetime.now(timezone.utc)
                write_issue_to_file(task, path)
                print(f'updated  [task    ]  "{task.title}"')
        return existing

    title = f"{repo}:{target_key}"
    request = IssueIdentifierRequest(
        title=title,
        existing_ids=frozenset(all_existing),
        prefix=project_key,
    )
    task_id = generate_issue_identifier(request).identifier
    all_existing.add(task_id)

    now = datetime.now(timezone.utc)
    task = IssueData.model_validate(
        {
            "id": task_id,
            "title": title,
            "description": f"Dependabot alerts for `{target_key}`.",
            "type": "task",
            "status": "open",
            "priority": priority,
            "assignee": None,
            "creator": None,
            "parent": parent_epic,
            "labels": ["security", "github", "dependabot"],
            "dependencies": [],
            "comments": [],
            "created_at": now.isoformat(),
            "updated_at": now.isoformat(),
            "closed_at": None,
            "custom": {
                "github_provider": "dependabot",
                "github_repository": repo,
                "github_manifest_path": target_key,
            },
        }
    )
    print(f'created  [task    ]  "{title}"')
    if not dry_run:
        write_issue_to_file(task, _issue_path(issues_dir, task_id))
    return task_id


def _map_dependabot_to_kanbus(
    alert: Dict[str, Any], repo: str, task_id: str
) -> IssueData:
    number = _alert_number(alert)
    package = _alert_package(alert)
    severity = _alert_severity(alert)
    state = _alert_state(alert)
    manifest_path = _alert_manifest(alert)
    ecosystem = _alert_ecosystem(alert)
    advisory = alert.get("security_advisory") or {}
    summary = advisory.get("summary") or "Dependabot alert"
    description = advisory.get("description") or ""
    ghsa_id = advisory.get("ghsa_id") or "dependabot-alert"
    html_url = alert.get("html_url") or ""

    title = (
        f"[Dependabot] {ghsa_id} in {package}"
        if package
        else f"[Dependabot] Alert #{number}: {summary}"
    )

    body = (
        f"## {summary}\n\n"
        f"**Provider:** GitHub Dependabot\n"
        f"**Repository:** `{repo}`\n"
        f"**Alert Number:** {number}\n"
        f"**Severity:** {severity}\n"
        f"**State:** {state}\n"
        f"**Package:** `{package}`\n"
        f"**Ecosystem:** `{ecosystem}`\n"
        f"**Manifest:** `{manifest_path}`\n\n"
        f"### Advisory\n{description}\n\n"
        f"### Reference\n- {html_url}"
    )

    now = datetime.now(timezone.utc)
    return IssueData.model_validate(
        {
            "id": "__placeholder__",
            "title": title,
            "description": body,
            "type": "sub-task",
            "status": "open",
            "priority": _severity_to_priority(severity),
            "assignee": None,
            "creator": None,
            "parent": task_id,
            "labels": ["security", "github", "dependabot"],
            "dependencies": [],
            "comments": [],
            "created_at": now.isoformat(),
            "updated_at": now.isoformat(),
            "closed_at": None,
            "custom": {
                "github_provider": "dependabot",
                "github_alert_number": number,
                "github_repository": repo,
                "github_manifest_path": manifest_path,
                "github_ecosystem": ecosystem,
                "github_package": package,
                "github_severity": severity,
                "github_alert_state": state,
                "github_html_url": html_url,
            },
        }
    )


def _build_alert_index(existing_ids: Set[str], issues_dir: Path) -> Dict[str, str]:
    index: Dict[str, str] = {}
    for identifier in existing_ids:
        try:
            issue = read_issue_from_file(_issue_path(issues_dir, identifier))
        except Exception:
            continue
        custom = issue.custom
        provider = custom.get("github_provider")
        number = custom.get("github_alert_number")
        repo = custom.get("github_repository")
        if (
            provider == "dependabot"
            and isinstance(number, int)
            and isinstance(repo, str)
        ):
            index[f"{repo}#{number}"] = identifier
    return index


def _build_task_index(existing_ids: Set[str], issues_dir: Path) -> Dict[str, str]:
    index: Dict[str, str] = {}
    for identifier in existing_ids:
        try:
            issue = read_issue_from_file(_issue_path(issues_dir, identifier))
        except Exception:
            continue
        if issue.issue_type != "task":
            continue
        custom = issue.custom
        if (
            custom.get("github_provider") == "dependabot"
            and isinstance(custom.get("github_manifest_path"), str)
            and custom.get("github_manifest_path")
        ):
            index[str(custom["github_manifest_path"])] = identifier
    return index


def _fetch_dependabot_alerts(repo: str, token: str, state: str) -> List[Dict[str, Any]]:
    headers = {
        "Authorization": f"Bearer {token}",
        "Accept": "application/vnd.github+json",
        "X-GitHub-Api-Version": GITHUB_API_VERSION,
        "User-Agent": "kanbus-cli",
    }
    alerts: List[Dict[str, Any]] = []
    url: Optional[str] = (
        f"{GITHUB_API_BASE}/repos/{repo}/dependabot/alerts"
        f"?state={state}&per_page=100"
    )
    while url:
        response = requests.get(url, headers=headers, timeout=30)
        if not response.ok:
            raise GithubSecuritySyncError(
                f"GitHub Dependabot API returned {response.status_code}: "
                f"{response.text[:400]}"
            )
        page = response.json()
        if not isinstance(page, list):
            raise GithubSecuritySyncError("unexpected Dependabot API response shape")
        alerts.extend(page)
        url = _parse_next_link(response.headers.get("Link"))
    return alerts


def _parse_next_link(link_header: Optional[str]) -> Optional[str]:
    if not link_header:
        return None
    for part in link_header.split(","):
        chunk = part.strip()
        if 'rel="next"' not in chunk:
            continue
        if "<" not in chunk or ">" not in chunk:
            return None
        return chunk[chunk.find("<") + 1 : chunk.find(">")]
    return None


def _task_target_key(alert: Dict[str, Any]) -> str:
    manifest = _alert_manifest(alert)
    if manifest and manifest != "unknown":
        return manifest
    ecosystem = _alert_ecosystem(alert)
    return ecosystem or "unknown"


def _min_priority(alerts: List[Dict[str, Any]]) -> int:
    return min(
        (_severity_to_priority(_alert_severity(item)) for item in alerts), default=3
    )


def _alert_number(alert: Dict[str, Any]) -> int:
    value = alert.get("number")
    return int(value) if isinstance(value, int) else 0


def _alert_state(alert: Dict[str, Any]) -> str:
    value = alert.get("state")
    return value if isinstance(value, str) else "open"


def _alert_severity(alert: Dict[str, Any]) -> str:
    advisory = alert.get("security_advisory") or {}
    sev = advisory.get("severity")
    if isinstance(sev, str):
        return sev
    vulnerability = alert.get("security_vulnerability") or {}
    sev = vulnerability.get("severity")
    return sev if isinstance(sev, str) else "low"


def _alert_manifest(alert: Dict[str, Any]) -> str:
    dependency = alert.get("dependency") or {}
    value = dependency.get("manifest_path")
    return value if isinstance(value, str) else "unknown"


def _alert_ecosystem(alert: Dict[str, Any]) -> str:
    dependency = alert.get("dependency") or {}
    package = dependency.get("package") or {}
    value = package.get("ecosystem")
    if isinstance(value, str):
        return value
    vuln = alert.get("security_vulnerability") or {}
    package = vuln.get("package") or {}
    value = package.get("ecosystem")
    return value if isinstance(value, str) else "unknown"


def _alert_package(alert: Dict[str, Any]) -> str:
    dependency = alert.get("dependency") or {}
    package = dependency.get("package") or {}
    value = package.get("name")
    if isinstance(value, str):
        return value
    vuln = alert.get("security_vulnerability") or {}
    package = vuln.get("package") or {}
    value = package.get("name")
    return value if isinstance(value, str) else "unknown"


def _dependabot_alert_title(alert: Dict[str, Any]) -> str:
    advisory = (alert.get("security_advisory") or {}).get(
        "ghsa_id"
    ) or "dependabot-alert"
    return f"[Dependabot] {advisory} in {_alert_package(alert)} #{_alert_number(alert)}"


def _severity_to_priority(severity: str) -> int:
    normalized = severity.lower()
    if normalized == "critical":
        return 0
    if normalized == "high":
        return 1
    if normalized == "medium":
        return 2
    return 3


def _detect_repo_from_git(root: Path) -> Optional[str]:
    try:
        result = subprocess.run(
            ["git", "remote", "get-url", "origin"],
            cwd=root,
            capture_output=True,
            text=True,
            timeout=5,
        )
    except Exception:
        return None
    return _extract_repo_slug(result.stdout.strip())


def _extract_repo_slug(remote: str) -> Optional[str]:
    if remote.startswith("https://github.com/"):
        return remote[len("https://github.com/") :].removesuffix(".git")
    if remote.startswith("git@github.com:"):
        return remote[len("git@github.com:") :].removesuffix(".git")
    return None


def _issue_path(issues_dir: Path, identifier: str) -> Path:
    return issues_dir / f"{identifier}.json"


def _metadata_marker_alert(repository: str, alert_number: int) -> str:
    return f"kanbus-gh-alert:dependabot|{repository}|{alert_number}"


def _metadata_marker_target(repository: str, target_key: str) -> str:
    return f"kanbus-gh-target:dependabot|{repository}|{target_key}"


def _append_marker(description: str, marker: str) -> str:
    return f"{description}\n\n<!-- {marker} -->"


def _extract_marker(description: str, prefix: str) -> Optional[str]:
    for line in description.splitlines():
        stripped = line.strip()
        if stripped.startswith("<!-- ") and stripped.endswith(" -->"):
            inner = stripped.removeprefix("<!-- ").removesuffix(" -->").strip()
            if inner.startswith(prefix):
                return inner[len(prefix) :]
    return None


def _build_beads_alert_index(issues: List[IssueData]) -> Dict[str, str]:
    index: Dict[str, str] = {}
    for issue in issues:
        marker = _extract_marker(issue.description or "", "kanbus-gh-alert:dependabot|")
        if not marker:
            continue
        parts = marker.split("|", 1)
        if len(parts) != 2:
            continue
        index[f"{parts[0]}#{parts[1]}"] = issue.identifier
    return index


def _build_beads_task_index(issues: List[IssueData]) -> Dict[str, str]:
    index: Dict[str, str] = {}
    for issue in issues:
        if issue.issue_type != "task":
            continue
        marker = _extract_marker(
            issue.description or "", "kanbus-gh-target:dependabot|"
        )
        if not marker:
            continue
        parts = marker.split("|", 1)
        if len(parts) != 2:
            continue
        index[parts[1]] = issue.identifier
    return index


def _resolve_beads_initiative(
    root: Path, issues: List[IssueData], dry_run: bool
) -> str:
    existing = _latest_beads_match(
        issues,
        lambda issue: (
            issue.issue_type == "initiative"
            and issue.title == GITHUB_SECURITY_INITIATIVE_TITLE
            and "github" in issue.labels
        ),
    ) or _latest_beads_match(
        issues,
        lambda issue: (
            issue.issue_type == "initiative"
            and issue.title == GITHUB_SECURITY_INITIATIVE_TITLE
        ),
    )
    if existing:
        return existing
    print(f'created  [initiative]  "{GITHUB_SECURITY_INITIATIVE_TITLE}"')
    if dry_run:
        return "would-create-initiative"
    created = create_beads_issue(
        root=root,
        title=GITHUB_SECURITY_INITIATIVE_TITLE,
        issue_type="initiative",
        priority=1,
        assignee=None,
        parent=None,
        description="Track remediation of GitHub security findings.",
    )
    update_beads_issue(
        root=root,
        identifier=created.identifier,
        status="open",
        title=GITHUB_SECURITY_INITIATIVE_TITLE,
        description="Track remediation of GitHub security findings.",
        priority=1,
        set_labels=["security", "github"],
    )
    return created.identifier


def _resolve_beads_epic(
    root: Path,
    issues: List[IssueData],
    configured_id: Optional[str],
    initiative_id: str,
    dry_run: bool,
) -> str:
    if configured_id and any(issue.identifier == configured_id for issue in issues):
        return configured_id
    existing = _latest_beads_match(
        issues,
        lambda issue: (
            issue.issue_type == "epic"
            and issue.title == GITHUB_DEPENDABOT_EPIC_TITLE
            and "dependabot" in issue.labels
        ),
    ) or _latest_beads_match(
        issues,
        lambda issue: (
            issue.issue_type == "epic" and issue.title == GITHUB_DEPENDABOT_EPIC_TITLE
        ),
    )
    if existing:
        return existing
    print(f'created  [epic    ]  "{GITHUB_DEPENDABOT_EPIC_TITLE}"')
    if dry_run:
        return "would-create-epic"
    created = create_beads_issue(
        root=root,
        title=GITHUB_DEPENDABOT_EPIC_TITLE,
        issue_type="epic",
        priority=1,
        assignee=None,
        parent=initiative_id,
        description="Dependabot alerts imported from GitHub Security.",
    )
    update_beads_issue(
        root=root,
        identifier=created.identifier,
        status="open",
        title=GITHUB_DEPENDABOT_EPIC_TITLE,
        description="Dependabot alerts imported from GitHub Security.",
        priority=1,
        set_labels=["security", "github", "dependabot"],
    )
    return created.identifier


def _latest_beads_match(
    issues: List[IssueData], predicate: Callable[[IssueData], bool]
) -> Optional[str]:
    matches = [issue for issue in issues if predicate(issue)]
    if not matches:
        return None
    matches.sort(key=lambda issue: issue.updated_at)
    return matches[-1].identifier


def _resolve_beads_task(
    root: Path,
    repository: str,
    target_key: str,
    parent_epic: str,
    priority: int,
    dry_run: bool,
    task_index: Dict[str, str],
) -> str:
    title = f"{repository}:{target_key}"
    description = _append_marker(
        f"Dependabot alerts for `{target_key}`.",
        _metadata_marker_target(repository, target_key),
    )

    existing = task_index.get(target_key)
    if existing:
        if not dry_run:
            update_beads_issue(
                root=root,
                identifier=existing,
                status="open",
                title=title,
                description=description,
                priority=priority,
                set_labels=["security", "github", "dependabot"],
            )
        return existing

    print(f'created  [task    ]  "{title}"')
    if dry_run:
        synthetic = f"would-create-task-{target_key}"
        task_index[target_key] = synthetic
        return synthetic

    created = create_beads_issue(
        root=root,
        title=title,
        issue_type="task",
        priority=priority,
        assignee=None,
        parent=parent_epic,
        description=description,
    )
    update_beads_issue(
        root=root,
        identifier=created.identifier,
        status="open",
        title=title,
        description=description,
        priority=priority,
        set_labels=["security", "github", "dependabot"],
    )
    task_index[target_key] = created.identifier
    return created.identifier


def _map_dependabot_to_beads_description(alert: Dict[str, Any], repo: str) -> str:
    number = _alert_number(alert)
    package = _alert_package(alert)
    severity = _alert_severity(alert)
    state = _alert_state(alert)
    manifest_path = _alert_manifest(alert)
    ecosystem = _alert_ecosystem(alert)
    advisory = alert.get("security_advisory") or {}
    summary = advisory.get("summary") or "Dependabot alert"
    description = advisory.get("description") or ""
    html_url = alert.get("html_url") or ""
    body = (
        f"## {summary}\n\n"
        f"**Provider:** GitHub Dependabot\n"
        f"**Repository:** `{repo}`\n"
        f"**Alert Number:** {number}\n"
        f"**Severity:** {severity}\n"
        f"**State:** {state}\n"
        f"**Package:** `{package}`\n"
        f"**Ecosystem:** `{ecosystem}`\n"
        f"**Manifest:** `{manifest_path}`\n\n"
        f"### Advisory\n{description}\n\n"
        f"### Reference\n- {html_url}"
    )
    return _append_marker(body, _metadata_marker_alert(repo, number))



from __future__ import annotations

import json
from pathlib import Path
from types import SimpleNamespace

import pytest

import kanbus.issue_delete as issue_delete
from test_helpers import build_issue


def _write_issue(project_dir: Path, issue_id: str, parent: str | None = None) -> None:
    issues_dir = project_dir / "issues"
    issues_dir.mkdir(parents=True, exist_ok=True)
    issue = build_issue(issue_id, parent=parent)
    payload = issue.model_dump(by_alias=True, mode="json")
    (issues_dir / f"{issue_id}.json").write_text(json.dumps(payload), encoding="utf-8")


def test_get_descendant_identifiers_includes_shared_and_local_leaf_first(
    tmp_path: Path,
) -> None:
    project_dir = tmp_path / "project"
    _write_issue(project_dir, "kanbus-root")
    _write_issue(project_dir, "kanbus-child", parent="kanbus-root")
    _write_issue(project_dir, "kanbus-leaf", parent="kanbus-child")

    local_project = tmp_path / "project-local"
    _write_issue(local_project, "kanbus-local-child", parent="kanbus-root")

    descendants = issue_delete.get_descendant_identifiers(project_dir, "kanbus-root")

    assert descendants == ["kanbus-leaf", "kanbus-child", "kanbus-local-child"]


def test_delete_issue_restores_file_when_event_delete_fails(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    project_dir = tmp_path / "project"
    issues_dir = project_dir / "issues"
    issues_dir.mkdir(parents=True, exist_ok=True)
    issue_path = issues_dir / "kanbus-1.json"
    issue_path.write_text('{"id": "kanbus-1"}\n', encoding="utf-8")
    issue = build_issue("kanbus-1")
    lookup = SimpleNamespace(
        issue=issue, issue_path=issue_path, project_dir=project_dir
    )

    monkeypatch.setattr(issue_delete, "load_issue_from_project", lambda _r, _i: lookup)
    monkeypatch.setattr(
        issue_delete,
        "delete_events_for_issues",
        lambda _events_dir, _ids: (_ for _ in ()).throw(
            RuntimeError("event delete failed")
        ),
    )

    with pytest.raises(issue_delete.IssueDeleteError, match="event delete failed"):
        issue_delete.delete_issue(tmp_path, "kanbus-1")

    assert issue_path.exists()
    restored = json.loads(issue_path.read_text(encoding="utf-8"))
    assert restored["id"] == "kanbus-1"


def test_delete_issue_publishes_deleted_for_shared_issue_path(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    project_dir = tmp_path / "project"
    issues_dir = project_dir / "issues"
    issues_dir.mkdir(parents=True, exist_ok=True)
    issue_path = issues_dir / "kanbus-2.json"
    issue_path.write_text('{"id": "kanbus-2"}\n', encoding="utf-8")
    issue = build_issue("kanbus-2")
    lookup = SimpleNamespace(
        issue=issue, issue_path=issue_path, project_dir=project_dir
    )
    published: list[tuple[Path, Path, str, None]] = []

    monkeypatch.setattr(issue_delete, "load_issue_from_project", lambda _r, _i: lookup)
    monkeypatch.setattr(
        issue_delete, "delete_events_for_issues", lambda _events_dir, _ids: None
    )
    monkeypatch.setattr(
        issue_delete,
        "publish_issue_deleted",
        lambda root, project, issue_id, event_id: published.append(
            (root, project, issue_id, event_id)
        ),
    )

    issue_delete.delete_issue(tmp_path, "kanbus-2")

    assert not issue_path.exists()
    assert published == [(tmp_path, project_dir, "kanbus-2", None)]


def test_delete_issue_does_not_publish_for_local_issue_path(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    project_dir = tmp_path / "project"
    local_issues_dir = tmp_path / "project-local" / "issues"
    local_issues_dir.mkdir(parents=True, exist_ok=True)
    issue_path = local_issues_dir / "kanbus-3.json"
    issue_path.write_text('{"id": "kanbus-3"}\n', encoding="utf-8")
    issue = build_issue("kanbus-3")
    lookup = SimpleNamespace(
        issue=issue, issue_path=issue_path, project_dir=project_dir
    )
    published: list[tuple[Path, Path, str, None]] = []

    monkeypatch.setattr(issue_delete, "load_issue_from_project", lambda _r, _i: lookup)
    monkeypatch.setattr(
        issue_delete, "delete_events_for_issues", lambda _events_dir, _ids: None
    )
    monkeypatch.setattr(
        issue_delete,
        "publish_issue_deleted",
        lambda root, project, issue_id, event_id: published.append(
            (root, project, issue_id, event_id)
        ),
    )

    issue_delete.delete_issue(tmp_path, "kanbus-3")

    assert not issue_path.exists()
    assert published == []


def test_delete_issue_wraps_lookup_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(
        issue_delete,
        "load_issue_from_project",
        lambda _r, _i: (_ for _ in ()).throw(issue_delete.IssueLookupError("missing")),
    )
    with pytest.raises(issue_delete.IssueDeleteError, match="missing"):
        issue_delete.delete_issue(tmp_path, "kanbus-missing")
